# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for calculating properties of sources defined by an Aperture.
"""

import functools
import inspect
import warnings
from copy import deepcopy

import astropy.units as u
import numpy as np
from astropy.nddata import NDData, StdDevUncertainty
from astropy.stats import (SigmaClip, biweight_location, biweight_midvariance,
                           mad_std)
from astropy.utils import lazyproperty
from astropy.utils.exceptions import AstropyUserWarning

from photutils.aperture import Aperture, SkyAperture, region_to_aperture
from photutils.aperture.core import _aperture_metadata
from photutils.morphology import gini as gini_func
from photutils.utils._deprecation import (create_empty_deprecated_qtable,
                                          deprecated_getattr,
                                          deprecated_positional_kwargs)
from photutils.utils._misc import _get_meta
from photutils.utils._moments import _image_moments
from photutils.utils._quantity_helpers import process_quantities

__all__ = ['ApertureStats']


# Default table columns for `to_table()` output
DEFAULT_COLUMNS = ['id', 'x_centroid', 'y_centroid', 'sky_centroid',
                   'sum', 'sum_err', 'sum_aper_area', 'center_aper_area',
                   'min', 'max', 'mean', 'median', 'mode', 'std',
                   'mad_std', 'var', 'biweight_location',
                   'biweight_midvariance', 'fwhm', 'semimajor_axis',
                   'semiminor_axis', 'orientation', 'eccentricity']

# Remove in 4.0
_DEPRECATED_ATTRIBUTES: dict = {
    'covar_sigx2': 'covariance_xx',
    'covar_sigxy': 'covariance_xy',
    'covar_sigy2': 'covariance_yy',
    'cxx': 'ellipse_cxx',
    'cxy': 'ellipse_cxy',
    'cyy': 'ellipse_cyy',
    'data_sumcutout': 'data_sum_cutout',
    'error_sumcutout': 'error_sum_cutout',
    'get_id': 'select_id',
    'get_ids': 'select_ids',
    'semimajor_sigma': 'semimajor_axis',
    'semiminor_sigma': 'semiminor_axis',
    'xcentroid': 'x_centroid',
    'ycentroid': 'y_centroid',
}


def as_scalar(method):
    """
    Return a decorated method where it will always return a scalar value
    (instead of a length-1 tuple/list/array) if the class is scalar.

    Parameters
    ----------
    method : function
        The method to be decorated.

    Returns
    -------
    decorator : function
        The decorated method.
    """
    @functools.wraps(method)
    def _decorator(*args, **kwargs):
        result = method(*args, **kwargs)
        try:
            return (result[0] if args[0].isscalar and len(result) == 1
                    else result)
        except TypeError:  # if result has no len
            return result

    return _decorator


class ApertureStats:
    """
    Class to create a catalog of statistics for pixels within an
    aperture.

    Note that this class returns the statistics of the input
    ``data`` values within the aperture. It does not convert data
    in surface brightness units to flux or counts. Conversion from
    surface-brightness units should be performed before using this
    function.

    Parameters
    ----------
    data : 2D `~numpy.ndarray`, `~astropy.units.Quantity`, \
            `~astropy.nddata.NDData`
        The 2D array from which to calculate the source properties.
        For accurate source properties, ``data`` should be
        background-subtracted. Non-finite ``data`` values (NaN and inf)
        are automatically masked.

    aperture : `~photutils.aperture.Aperture` or supported `~regions.Region`
        The aperture or region to apply to the data. The aperture
        or region object may contain more than one position. If the
        input ``aperture`` is a `~photutils.aperture.SkyAperture` or
        `~regions.SkyRegion` object, then a WCS must be input using
        the ``wcs`` keyword. Region objects are converted to aperture
        objects.

    error : 2D `~numpy.ndarray` or `~astropy.units.Quantity`, optional
        The total error array corresponding to the input ``data``
        array. ``error`` is assumed to include *all* sources of
        error, including the Poisson error of the sources (see
        `~photutils.utils.calc_total_error`). ``error`` must have
        the same shape as the input ``data``. If ``data`` is a
        `~astropy.units.Quantity` array then ``error`` must be a
        `~astropy.units.Quantity` array (and vice versa) with identical
        units. Non-finite ``error`` values (NaN and +/- inf) are not
        automatically masked, unless they are at the same position of
        non-finite values in the input ``data`` array. Such pixels can
        be masked using the ``mask`` keyword.

    mask : 2D `~numpy.ndarray` (bool), optional
        A boolean mask with the same shape as ``data`` where a `True`
        value indicates the corresponding element of ``data`` is masked.
        Masked data are excluded from all calculations. Non-finite
        values (NaN and inf) in the input ``data`` are automatically
        masked.

    wcs : WCS object or `None`, optional
        A world coordinate system (WCS) transformation that
        supports the `astropy shared interface for WCS
        <https://docs.astropy.org/en/stable/wcs/wcsapi.html>`_ (e.g.,
        `astropy.wcs.WCS`, `gwcs.wcs.WCS`). ``wcs`` is required if
        the input ``aperture`` is a `~photutils.aperture.SkyAperture`
        or `~regions.SkyRegion` object. If `None`, then all sky-based
        properties will be set to `None`.

    sigma_clip : `None` or `astropy.stats.SigmaClip` instance, optional
        A `~astropy.stats.SigmaClip` object that defines the sigma
        clipping parameters. If `None` then no sigma clipping will
        be performed.

    sum_method : {'exact', 'center', 'subpixel'}, optional
        The method used to determine the overlap of the aperture on
        the pixel grid. This method is used only for calculating the
        ``sum``, ``sum_error``, ``sum_aper_area``, ``data_sum_cutout``,
        and ``error_sum_cutout`` properties. All other properties use
        the "center" aperture mask method. Not all options are available
        for all aperture types. The following methods are available:

        * ``'exact'`` (default):
          The exact fractional overlap of the aperture and each pixel is
          calculated. The aperture weights will contain values between 0
          and 1.

        * ``'center'``:
          A pixel is considered to be entirely in or out of the aperture
          depending on whether its center is in or out of the aperture.
          The aperture weights will contain values only of 0 (out) and 1
          (in).

        * ``'subpixel'``:
          A pixel is divided into subpixels (see the ``subpixels``
          keyword), each of which are considered to be entirely in or
          out of the aperture depending on whether its center is in
          or out of the aperture. If ``subpixels=1``, this method is
          equivalent to ``'center'``. The aperture weights will contain
          values between 0 and 1.

    subpixels : int, optional
        For the ``'subpixel'`` method, resample pixels by this factor
        in each dimension. That is, each pixel is divided into
        ``subpixels**2`` subpixels. This keyword is ignored unless
        ``sum_method='subpixel'``.

    local_bkg : float, `~numpy.ndarray`, `~astropy.units.Quantity`, or `None`
        The per-pixel local background values to subtract from the data
        before performing measurements. If input as an array, the order
        of ``local_bkg`` values corresponds to the order of the input
        ``aperture`` positions. ``local_bkg`` must have the same length
        as the input ``aperture`` or must be a scalar value, which
        will be broadcast to all apertures. If `None`, then no local
        background subtraction is performed. If the input ``data`` has
        units, then ``local_bkg`` must be a `~astropy.units.Quantity`
        with the same units.

    Notes
    -----
    ``data`` should be background-subtracted for accurate source
    properties. In addition to global background subtraction, local
    background subtraction can be performed using the ``local_bkg``
    keyword values.

    `~regions.Region` objects are converted to `Aperture` objects using
    the :func:`region_to_aperture` function.

    The returned statistics are measured for the pixels within the
    input aperture at its input position. This class does not change
    the position of the input aperture. This class returns the centroid
    value of the pixels within the input aperture, but the input
    aperture is not recentered at the measured centroid position
    when making the measurements. If desired, you can create a new
    `Aperture` object using the measured centroid and then re-run
    `~photutils.aperture.ApertureStats`.

    Most source properties are calculated using the "center"
    aperture-mask method, which gives aperture weights of 0 or 1. This
    avoids the need to compute weighted statistics --- the ``data``
    pixel values are directly used.

    The input ``sum_method`` and ``subpixels`` keywords are used
    to determine the aperture-mask method when calculating the
    sum-related properties: ``sum``, ``sum_error``, ``sum_aper_area``,
    ``data_sum_cutout``, and ``error_sum_cutout``. The default is
    ``sum_method='exact'``, which produces exact aperture-weighted
    photometry.

    .. _SourceExtractor: https://sextractor.readthedocs.io/en/latest/

    Examples
    --------
    >>> from photutils.datasets import make_4gaussians_image
    >>> from photutils.aperture import CircularAperture, ApertureStats

    >>> data = make_4gaussians_image()
    >>> aper = CircularAperture((150, 25), 8)
    >>> aperstats = ApertureStats(data, aper)
    >>> print(aperstats.x_centroid)  # doctest: +FLOAT_CMP
    149.99080259251238
    >>> print(aperstats.y_centroid)  # doctest: +FLOAT_CMP
    24.97484633000507
    >>> print(aperstats.centroid)  # doctest: +FLOAT_CMP
    [149.99080259  24.97484633]

    >>> print(aperstats.mean, aperstats.median)  # doctest: +FLOAT_CMP
    47.76300955780609 31.913789514433084

    >>> print(aperstats.std)  # doctest: +FLOAT_CMP
    39.193655383492974

    >>> print(aperstats.sum)  # doctest: +FLOAT_CMP
    9286.709206410273

    >>> print(aperstats.sum_aper_area) # doctest: +FLOAT_CMP
    201.0619298297468 pix2

    >>> # More than one aperture position
    >>> aper2 = CircularAperture(((150, 25), (90, 60)), 10)
    >>> aperstats2 = ApertureStats(data, aper2)
    >>> print(aperstats2.x_centroid)  # doctest: +FLOAT_CMP
    [149.98470724  89.97893946]
    >>> print(aperstats2.sum)  # doctest: +FLOAT_CMP
    [10177.62548482 36653.97704059]
    """

    def __init__(self, data, aperture, *, error=None, mask=None, wcs=None,
                 sigma_clip=None, sum_method='exact', subpixels=5,
                 local_bkg=None):

        if isinstance(data, NDData):
            data, error, mask, wcs = self._unpack_nddata(data, error, mask,
                                                         wcs)

        inputs = (data, error, local_bkg)
        names = ('data', 'error', 'local_bkg')
        inputs, unit = process_quantities(inputs, names)
        (data, error, local_bkg) = inputs

        self._data = self._validate_array(data, 'data', shape=False)
        self._data_unit = unit
        self._input_aperture = self._validate_aperture(aperture)
        aperture_meta = _aperture_metadata(aperture)  # use input aperture

        if isinstance(aperture, SkyAperture) and wcs is None:
            msg = 'A wcs is required when using a SkyAperture'
            raise ValueError(msg)

        # Convert region to aperture if necessary
        if not isinstance(aperture, Aperture):
            aperture = region_to_aperture(aperture)
        self.aperture = aperture

        self._error = self._validate_array(error, 'error')
        self._mask = self._validate_array(mask, 'mask')
        self._wcs = wcs

        if sigma_clip is not None and not isinstance(sigma_clip, SigmaClip):
            msg = 'sigma_clip must be a SigmaClip instance'
            raise TypeError(msg)
        self.sigma_clip = sigma_clip

        self.sum_method = sum_method
        self.subpixels = subpixels

        self._local_bkg = np.zeros(self.n_apertures)  # no local bkg
        if local_bkg is not None:
            local_bkg = np.atleast_1d(local_bkg)
            if local_bkg.ndim != 1:
                msg = 'local_bkg must be a 1D array'
                raise ValueError(msg)

            n_local_bkg = len(local_bkg)
            if n_local_bkg not in (1, self.n_apertures):
                msg = ('local_bkg must be scalar or have the same length '
                       'as the input aperture')
                raise ValueError(msg)
            local_bkg = np.broadcast_to(local_bkg, self.n_apertures)

            if np.any(~np.isfinite(local_bkg)):
                msg = ('local_bkg must not contain any non-finite '
                       '(e.g., inf or NaN) values')
                raise ValueError(msg)
            self._local_bkg = local_bkg  # always an iterable

        self._ids = np.arange(self.n_apertures) + 1
        self.default_columns = DEFAULT_COLUMNS
        self.meta = _get_meta()
        self.meta.update(aperture_meta)

    @staticmethod
    def _unpack_nddata(data, error, mask, wcs):
        nddata_attr = {'error': error, 'mask': mask, 'wcs': wcs}
        for key, value in nddata_attr.items():
            if value is not None:
                msg = (f'The {key!r} keyword will be ignored. Its value '
                       'is obtained from the input NDData object.')
                warnings.warn(msg, AstropyUserWarning)

        mask = data.mask
        wcs = data.wcs

        if isinstance(data.uncertainty, StdDevUncertainty):
            if data.uncertainty.unit is None:
                error = data.uncertainty.array
            else:
                error = data.uncertainty.array * data.uncertainty.unit

        if data.unit is not None:
            data = u.Quantity(data.data, unit=data.unit)
        else:
            data = data.data

        return data, error, mask, wcs

    @staticmethod
    def _validate_aperture(aperture):
        try:
            from regions import Region
            aper_types = (Aperture, Region)
        except ImportError:
            aper_types = Aperture

        if not isinstance(aperture, aper_types):
            msg = 'aperture must be an Aperture or Region object'
            raise TypeError(msg)
        return aperture

    def _validate_array(self, array, name, *, ndim=2, shape=True):
        if name == 'mask' and array is np.ma.nomask:
            array = None
        if array is not None:
            array = np.asanyarray(array)
            if array.ndim != ndim:
                msg = f'{name} must be a {ndim}D array'
                raise ValueError(msg)
            if shape and array.shape != self._data.shape:
                msg = f'data and {name} must have the same shape'
                raise ValueError(msg)
        return array

    @property
    def _lazyproperties(self):
        """
        A list of all class lazyproperties (even in superclasses).

        The result is cached on the class to avoid repeated
        introspection via `inspect.getmembers`.
        """
        cls = self.__class__
        attr = '_cached_lazyproperties'
        # Subclasses get their own lazyproperty list
        if attr not in cls.__dict__:
            def islazyproperty(obj):
                return isinstance(obj, lazyproperty)

            setattr(cls, attr,
                    [i[0] for i in inspect.getmembers(
                        cls, predicate=islazyproperty)])
        return getattr(cls, attr)

    @property
    def properties(self):
        """
        A sorted list of the built-in source properties.
        """
        lazyproperties = [name for name in self._lazyproperties if not
                          name.startswith('_')]
        lazyproperties.sort()
        return lazyproperties

    def __getitem__(self, index):
        if self.isscalar:
            msg = (f'A scalar {self.__class__.__name__!r} object cannot '
                   'be indexed')
            raise TypeError(msg)

        newcls = object.__new__(self.__class__)

        # Attributes defined in __init__ that are copied directly to the
        # new class
        init_attr = ('_data', '_data_unit', '_error', '_mask', '_wcs',
                     'sigma_clip', 'sum_method', 'subpixels',
                     'default_columns', 'meta')
        for attr in init_attr:
            setattr(newcls, attr, getattr(self, attr))

        # Need to slice _aperture and _ids;
        # aperture determines isscalar (needed below)
        attrs = ('aperture', '_ids')
        for attr in attrs:
            setattr(newcls, attr, getattr(self, attr)[index])

        # Slice evaluated lazyproperty objects
        keys = set(self.__dict__.keys()) & set(self._lazyproperties)
        keys.add('_local_bkg')  # iterable defined in __init__
        for key in keys:
            value = self.__dict__[key]

            # Do not insert attributes that are always scalar (e.g.,
            # isscalar, n_apertures), i.e., not an array/list for each
            # source
            if np.isscalar(value):
                continue

            try:
                # Keep most _<attrs> as length-1 iterables
                if (newcls.isscalar and key.startswith('_')
                        and key != '_pixel_aperture'):
                    if isinstance(value, np.ndarray):
                        val = value[:, np.newaxis][index]
                    else:
                        val = [value[index]]
                else:
                    val = value[index]
            except TypeError:
                # Apply fancy indices (e.g., array/list or bool mask) to
                # lists.
                # See https://numpy.org/doc/stable/release/1.20.0-notes.html
                # #arraylike-objects-which-do-not-define-len-and-getitem
                arr = np.empty(len(value), dtype=object)
                arr[:] = list(value)
                val = arr[index].tolist()

            newcls.__dict__[key] = val
        return newcls

    def __str__(self):
        cls_name = f'<{self.__class__.__module__}.{self.__class__.__name__}>'
        with np.printoptions(threshold=25, edgeitems=5):
            fmt = [f'Length: {self.n_apertures}']
        return f'{cls_name}\n' + '\n'.join(fmt)

    def __repr__(self):
        return self.__str__()

    def __len__(self):
        if self.isscalar:
            msg = f'Scalar {self.__class__.__name__!r} object has no len()'
            raise TypeError(msg)
        return self.n_apertures

    def __iter__(self):
        for item in range(len(self)):
            yield self.__getitem__(item)

    # Remove in 4.0
    def __getattr__(self, name):
        return deprecated_getattr(self, name, _DEPRECATED_ATTRIBUTES,
                                  since='3.0', until='4.0')

    @lazyproperty
    def isscalar(self):
        """
        Whether the instance is scalar (e.g., a single aperture
        position).
        """
        return self._pixel_aperture.isscalar

    def copy(self):
        """
        Return a deep copy of this object.

        Returns
        -------
        result : `ApertureStats`
            A deep copy of this object.
        """
        return deepcopy(self)

    @lazyproperty
    def _null_object(self):
        """
        Return `None` values.
        """
        return np.array([None] * self.n_apertures)

    @lazyproperty
    def _null_value(self):
        """
        Return np.nan values.
        """
        values = np.empty(self.n_apertures)
        values.fill(np.nan)
        return values

    @property
    @as_scalar
    def id(self):
        """
        The aperture identification number(s).
        """
        return self._ids

    @property
    def ids(self):
        """
        The aperture identification number(s), always as an iterable
        `~numpy.ndarray`.
        """
        _ids = self._ids
        if self.isscalar:
            _ids = np.array((_ids,))
        return _ids

    def select_id(self, id_num):
        """
        Return a new `ApertureStats` object for the input ID number
        only.

        Parameters
        ----------
        id_num : int
            The aperture ID number.

        Returns
        -------
        result : `ApertureStats`
            A new `ApertureStats` object containing only the source with
            the input ID number.
        """
        return self.select_ids(id_num)

    def select_ids(self, id_nums):
        """
        Return a new `ApertureStats` object for the input ID numbers
        only.

        Parameters
        ----------
        id_nums : list, tuple, or `~numpy.ndarray` of int
            The aperture ID number(s).

        Returns
        -------
        result : `ApertureStats`
            A new `ApertureStats` object containing only the sources with
            the input ID numbers.
        """
        for id_num in np.atleast_1d(id_nums):
            if id_num not in self.ids:
                msg = f'{id_num} is not a valid source ID number'
                raise ValueError(msg)

        sorter = np.argsort(self.id)
        indices = sorter[np.searchsorted(self.id, id_nums, sorter=sorter)]
        return self[indices]

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def to_table(self, *, columns=None):
        """
        Create a `~astropy.table.QTable` of source properties.

        Parameters
        ----------
        columns : str, list of str, `None`, optional
            Names of columns, in order, to include in the output
            `~astropy.table.QTable`. The allowed column names are any of
            the `ApertureStats` properties. If ``columns`` is `None`,
            then a default list of scalar-valued properties (as defined
            by the ``default_columns`` attribute) will be used.

        Returns
        -------
        table : `~astropy.table.QTable`
            A table of sources properties with one row per source.
        """
        if columns is None:
            table_columns = self.default_columns
        elif isinstance(columns, str):
            table_columns = [columns]
        else:
            table_columns = columns

        # Replace with QTable in 4.0
        tbl = create_empty_deprecated_qtable(
            _DEPRECATED_ATTRIBUTES, since='3.0', until='4.0')

        tbl.meta.update(self.meta)  # keep tbl.meta type

        for column in table_columns:
            values = getattr(self, column)

            # Column assignment requires an object with a length
            if self.isscalar:
                values = (values,)

            tbl[column] = values
        return tbl

    @lazyproperty
    def n_apertures(self):
        """
        The number of positions for the input aperture.
        """
        if self.isscalar:
            return 1
        return len(self._pixel_aperture)

    @lazyproperty
    def _pixel_aperture(self):
        """
        The input aperture as a PixelAperture.
        """
        if isinstance(self.aperture, SkyAperture):
            return self.aperture.to_pixel(self._wcs)
        return self.aperture

    @lazyproperty
    def _aperture_masks_center(self):
        """
        The aperture masks (`ApertureMask`) generated with the 'center'
        method, always as an iterable.
        """
        aperture_masks = self._pixel_aperture.to_mask(method='center')
        if self.isscalar:
            aperture_masks = (aperture_masks,)
        return aperture_masks

    @lazyproperty
    def _aperture_masks(self):
        """
        The aperture masks (`ApertureMask`) generated with the
        ``sum_method`` method, always as an iterable.
        """
        aperture_masks = self._pixel_aperture.to_mask(method=self.sum_method,
                                                      subpixels=self.subpixels)
        if self.isscalar:
            aperture_masks = (aperture_masks,)
        return aperture_masks

    @lazyproperty
    def _overlap_slices(self):
        """
        The aperture mask overlap slices with the data, always as an
        iterable.

        The overlap slices are the same for all aperture mask methods.
        """
        overlap_slices = []
        for apermask in self._aperture_masks_center:
            (slc_large, slc_small) = apermask.get_overlap_slices(
                self._data.shape)
            overlap_slices.append((slc_large, slc_small))
        return overlap_slices

    @lazyproperty
    def _data_cutouts(self):
        """
        The local-background-subtracted unmasked data cutouts using the
        aperture bounding box, always as an iterable.
        """
        cutouts = []
        for (slices, local_bkg) in zip(self._overlap_slices,
                                       self._local_bkg, strict=True):
            if slices[0] is None:
                cutout = None  # no aperture overlap with the data
            else:
                # Copy is needed to preserve input data because masks are
                # applied to these cutouts later
                cutout = (self._data[slices[0]].astype(float, copy=True)
                          - local_bkg)
            cutouts.append(cutout)
        return cutouts

    def _make_aperture_cutouts(self, aperture_masks):
        """
        Make aperture-weighted cutouts for the data and variance, and
        cutouts for the total mask and aperture mask weights.

        Parameters
        ----------
        aperture_masks : list of `ApertureMask`
            A list of `ApertureMask` objects.

        Returns
        -------
        data, variance, mask, weights : list of `~numpy.ndarray`
            A list of cutout arrays for the data, variance, mask and weight
            arrays for each source (aperture position).
        """
        data_cutouts = []
        variance_cutouts = []
        mask_cutouts = []
        weight_cutouts = []
        overlaps = []

        for (data_cutout, apermask, slices) in zip(self._data_cutouts,
                                                   aperture_masks,
                                                   self._overlap_slices,
                                                   strict=True):

            slc_large, slc_small = slices
            if slc_large is None:  # aperture does not overlap the data
                overlap = False
                data_cutout = np.array([np.nan])
                variance_cutout = np.array([np.nan])
                mask_cutout = np.array([False])
                weight_cutout = np.array([np.nan])
            else:
                # Create a mask of non-finite ``data`` values combined
                # with the input ``mask`` array
                data_mask = ~np.isfinite(data_cutout)
                if self._mask is not None:
                    data_mask |= self._mask[slc_large]

                overlap = True
                aperweight_cutout = apermask.data[slc_small]
                weight_cutout = aperweight_cutout * ~data_mask

                # Apply the aperture mask; for "exact" and "subpixel"
                # this is an expanded boolean mask using the aperture
                # mask zero values
                mask_cutout = (aperweight_cutout == 0) | data_mask

                data_cutout = data_cutout.copy()
                if self.sigma_clip is None:
                    # data_cutout will have zeros where mask_cutout is True
                    data_cutout *= ~mask_cutout
                else:
                    # To input a mask, SigmaClip needs a MaskedArray
                    data_cutout_ma = np.ma.masked_array(data_cutout,
                                                        mask=mask_cutout)
                    data_sigclip = self.sigma_clip(data_cutout_ma)

                    # Define a mask of only the sigma-clipped pixels
                    sigclip_mask = data_sigclip.mask & ~mask_cutout
                    weight_cutout *= ~sigclip_mask

                    mask_cutout = data_sigclip.mask
                    data_cutout = data_sigclip.filled(0.0)

                # Need to apply the aperture weights
                data_cutout *= aperweight_cutout

                if self._error is None:
                    variance_cutout = None
                else:
                    # Apply the exact weights and total mask;
                    # error_cutout will have zeros where mask_cutout is True
                    variance = self._error[slc_large]**2
                    variance_cutout = (variance * aperweight_cutout
                                       * ~mask_cutout)

            data_cutouts.append(data_cutout)
            variance_cutouts.append(variance_cutout)
            mask_cutouts.append(mask_cutout)
            weight_cutouts.append(weight_cutout)
            overlaps.append(overlap)

        # Use zip (instead of np.transpose) because these may contain
        # arrays that have different shapes
        return list(zip(data_cutouts, variance_cutouts, mask_cutouts,
                        weight_cutouts, overlaps, strict=True))

    @lazyproperty
    def _aperture_cutouts_center(self):
        """
        Aperture-weighted cutouts for the data, variance, total mask,
        and aperture weights using the "center" aperture mask method.
        """
        return self._make_aperture_cutouts(self._aperture_masks_center)

    @lazyproperty
    def _aperture_cutouts(self):
        """
        Aperture-weighted cutouts for the data, variance, total mask,
        and aperture weights using the input ``sum_method`` aperture
        mask method.
        """
        return self._make_aperture_cutouts(self._aperture_masks)

    @lazyproperty
    def _mask_cutout_center(self):
        """
        Boolean mask cutouts representing the total mask.

        The total mask is combination of the input ``mask``, non-finite
        ``data`` values, the cutout aperture mask using the "center"
        method, and the sigma-clip mask.
        """
        return list(zip(*self._aperture_cutouts_center, strict=True))[2]

    @lazyproperty
    def _mask_cutout(self):
        """
        Boolean mask cutouts representing the total mask.

        The total mask is combination of the input ``mask``,
        non-finite ``data`` values, the cutout aperture mask using the
        ``sum_method`` method, and the sigma-clip mask.
        """
        return list(zip(*self._aperture_cutouts, strict=True))[2]

    def _make_masked_array_center(self, array):
        """
        Return a list of cutout masked arrays using the ``_mask_cutout``
        mask.

        Units are not applied.
        """
        return [np.ma.masked_array(arr, mask=mask)
                for arr, mask in zip(array, self._mask_cutout_center,
                                     strict=True)]

    def _make_masked_array(self, array):
        """
        Return a list of cutout masked arrays using the
        ``_mask_sumcutout`` mask.

        Units are not applied.
        """
        return [np.ma.masked_array(arr, mask=mask)
                for arr, mask in zip(array, self._mask_cutout, strict=True)]

    @lazyproperty
    @as_scalar
    def data_cutout(self):
        """
        A 2D aperture-weighted cutout from the data using the aperture
        mask with the "center" method as a `~numpy.ma.MaskedArray`.

        The cutout does not have units due to current limitations of
        masked quantity arrays.

        The mask is `True` for pixels from the input ``mask``,
        non-finite ``data`` values (NaN and inf), sigma-clipped pixels
        within the aperture, and pixels where the aperture mask has zero
        weight.
        """
        return self._make_masked_array_center(
            list(zip(*self._aperture_cutouts_center, strict=True))[0])

    @lazyproperty
    @as_scalar
    def data_sum_cutout(self):
        """
        A 2D aperture-weighted cutout from the data using the aperture
        mask with the input ``sum_method`` method as a
        `~numpy.ma.MaskedArray`.

        The cutout does not have units due to current limitations of
        masked quantity arrays.

        The mask is `True` for pixels from the input ``mask``,
        non-finite ``data`` values (NaN and inf), sigma-clipped pixels
        within the aperture, and pixels where the aperture mask has zero
        weight.
        """
        return self._make_masked_array(list(zip(*self._aperture_cutouts,
                                                strict=True))[0])

    @lazyproperty
    def _variance_cutout_center(self):
        """
        A 2D aperture-weighted variance cutout using the aperture mask
        with the input "center" method as a `~numpy.ma.MaskedArray`.

        The cutout does not have units due to current limitations of
        masked quantity arrays.

        The mask is `True` for pixels from the input ``mask``,
        non-finite ``data`` values (NaN and inf), sigma-clipped pixels
        within the aperture, and pixels where the aperture mask has zero
        weight.
        """
        if self._error is None:
            return self._null_object
        return self._make_masked_array_center(
            list(zip(*self._aperture_cutouts_center, strict=True))[1])

    @lazyproperty
    def _variance_cutout(self):
        """
        A 2D aperture-weighted variance cutout using the aperture mask
        with the input ``sum_method`` method as a
        `~numpy.ma.MaskedArray`.

        The cutout does not have units due to current limitations of
        masked quantity arrays.

        The mask is `True` for pixels from the input ``mask``,
        non-finite ``data`` values (NaN and inf), sigma-clipped pixels
        within the aperture, and pixels where the aperture mask has zero
        weight.
        """
        if self._error is None:
            return self._null_object
        return self._make_masked_array(list(zip(*self._aperture_cutouts,
                                                strict=True))[1])

    @lazyproperty
    @as_scalar
    def error_sum_cutout(self):
        """
        A 2D aperture-weighted error cutout using the aperture mask with
        the input ``sum_method`` method as a `~numpy.ma.MaskedArray`.

        The cutout does not have units due to current limitations of
        masked quantity arrays.

        The mask is `True` for pixels from the input ``mask``,
        non-finite ``data`` values (NaN and inf), sigma-clipped pixels
        within the aperture, and pixels where the aperture mask has zero
        weight.
        """
        if self._error is None:
            return self._null_object
        return [np.sqrt(var) for var in self._variance_cutout]

    @lazyproperty
    def _weight_cutout_center(self):
        """
        A 2D `~numpy.ma.MaskedArray` cutout from the aperture mask
        weights array using the aperture bounding box.

        The aperture mask weights are for the "center" method.

        The mask is `True` for pixels outside the aperture mask, pixels
        from the input ``mask``, non-finite ``data`` values (NaN and
        inf), and sigma-clipped pixels.
        """
        return self._make_masked_array_center(
            list(zip(*self._aperture_cutouts_center, strict=True))[3])

    @lazyproperty
    def _weight_cutout(self):
        """
        A 2D `~numpy.ma.MaskedArray` cutout from the aperture mask
        weights array using the aperture bounding box.

        The aperture mask weights are for the ``sum_method`` method.

        The mask is `True` for pixels outside the aperture mask, pixels
        from the input ``mask``, non-finite ``data`` values (NaN and
        inf), and sigma-clipped pixels.
        """
        return self._make_masked_array(list(zip(*self._aperture_cutouts,
                                                strict=True))[3])

    @lazyproperty
    def _moment_data_cutout(self):
        """
        A list of 2D `~numpy.ndarray` cutouts from the data.

        Masked pixels are set to zero in these arrays (zeros do not
        contribute to the image moments). The aperture mask weights are
        for the "center" method.

        These arrays are used to derive moment-based properties.
        """
        data = deepcopy(self.data_cutout)  # self.data_cutout is a list
        if self.isscalar:
            data = (data,)

        cutouts = []
        for arr in data:
            if arr.size == 1 and np.isnan(arr[0]):  # no aperture overlap
                arr_ = np.empty((2, 2))
                arr_.fill(np.nan)
            else:
                arr_ = arr.data
                arr_[arr.mask] = 0.0
            cutouts.append(arr_)

        return cutouts

    @lazyproperty
    def _all_masked(self):
        """
        True if all pixels within the aperture are masked.
        """
        return np.array([np.all(mask) for mask in self._mask_cutout_center])

    @lazyproperty
    def _overlap(self):
        """
        True if there is no overlap of the aperture with the data.
        """
        return list(zip(*self._aperture_cutouts_center, strict=True))[4]

    def _get_values(self, array):
        """
        Get a 1D array of unmasked aperture-weighted values from the
        input array.

        An array with a single NaN is returned for completely-masked
        sources.
        """
        if self.isscalar:
            array = (array,)
        return [arr.compressed() if len(arr.compressed()) > 0
                else np.array([np.nan]) for arr in array]

    @lazyproperty
    def _data_values_center(self):
        """
        A 1D array of unmasked aperture-weighted data values using the
        "center" method.

        An array with a single NaN is returned for completely-masked
        sources.
        """
        return self._get_values(self.data_cutout)

    @lazyproperty
    @as_scalar
    def moments(self):
        """
        Spatial moments up to 3rd order of the source.
        """
        return np.array([_image_moments(arr, order=3)
                         for arr in self._moment_data_cutout])

    @lazyproperty
    @as_scalar
    def moments_central(self):
        """
        Central moments (translation invariant) of the source up to 3rd
        order.
        """
        cutout_centroid = self.cutout_centroid
        if self.isscalar:
            cutout_centroid = cutout_centroid[np.newaxis, :]
        return np.array([_image_moments(arr, center=(xcen_, ycen_), order=3)
                         for arr, xcen_, ycen_ in
                         zip(self._moment_data_cutout, cutout_centroid[:, 0],
                             cutout_centroid[:, 1], strict=True)])

    @lazyproperty
    @as_scalar
    def cutout_centroid(self):
        """
        The ``(x, y)`` coordinate, relative to the cutout data, of the
        centroid within the aperture.

        The centroid is computed as the center of mass of the unmasked
        pixels within the aperture.
        """
        moments = self.moments
        if self.isscalar:
            moments = moments[np.newaxis, :]

        # Ignore divide-by-zero RuntimeWarning
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            y_centroid = moments[:, 1, 0] / moments[:, 0, 0]
            x_centroid = moments[:, 0, 1] / moments[:, 0, 0]
        return np.transpose((x_centroid, y_centroid))

    @lazyproperty
    @as_scalar
    def centroid(self):
        """
        The ``(x, y)`` coordinate of the centroid.

        The centroid is computed as the center of mass of the unmasked
        pixels within the aperture.
        """
        origin = np.transpose((self.bbox_xmin, self.bbox_ymin))
        return self.cutout_centroid + origin

    @lazyproperty
    def _x_centroid(self):
        """
        The ``x`` coordinate of the centroid, always as an iterable.
        """
        x_centroid = np.transpose(self.centroid)[0]
        if self.isscalar:
            x_centroid = (x_centroid,)
        return x_centroid

    @lazyproperty
    @as_scalar
    def x_centroid(self):
        """
        The ``x`` coordinate of the centroid.

        The centroid is computed as the center of mass of the unmasked
        pixels within the aperture.
        """
        return self._x_centroid

    @lazyproperty
    def _y_centroid(self):
        """
        The ``y`` coordinate of the centroid, always as an iterable.
        """
        y_centroid = np.transpose(self.centroid)[1]
        if self.isscalar:
            y_centroid = (y_centroid,)
        return y_centroid

    @lazyproperty
    @as_scalar
    def y_centroid(self):
        """
        The ``y`` coordinate of the centroid.

        The centroid is computed as the center of mass of the unmasked
        pixels within the aperture.
        """
        return self._y_centroid

    @lazyproperty
    @as_scalar
    def sky_centroid(self):
        """
        The sky coordinate of the centroid of the unmasked pixels within
        the aperture, returned as a `~astropy.coordinates.SkyCoord`
        object.

        The output coordinate frame is the same as the input ``wcs``.

        `None` if ``wcs`` is not input.
        """
        if self._wcs is None:
            return self._null_object
        return self._wcs.pixel_to_world(self.x_centroid, self.y_centroid)

    @lazyproperty
    @as_scalar
    def sky_centroid_icrs(self):
        """
        The sky coordinate in the International Celestial Reference
        System (ICRS) frame of the centroid of the unmasked pixels
        within the aperture, returned as a
        `~astropy.coordinates.SkyCoord` object.

        `None` if ``wcs`` is not input.
        """
        if self._wcs is None:
            return self._null_object
        return self.sky_centroid.icrs

    @lazyproperty
    def _bbox(self):
        """
        The `~photutils.aperture.BoundingBox` of the aperture, always as
        an iterable.
        """
        apertures = self._pixel_aperture
        if self.isscalar:
            apertures = (apertures,)
        return [aperture.bbox for aperture in apertures]

    @lazyproperty
    @as_scalar
    def bbox(self):
        """
        The `~photutils.aperture.BoundingBox` of the aperture.

        Note that the aperture bounding box is calculated using the
        exact size of the aperture, which may be slightly larger than
        the aperture mask calculated using the "center" method.
        """
        return self._bbox

    @lazyproperty
    @as_scalar
    def _bbox_bounds(self):
        """
        The bounding box x and y minimum and maximum bounds.
        """
        bbox = self.bbox
        if self.isscalar:
            bbox = (bbox,)
        return np.array([(bbox_.ixmin, bbox_.ixmax - 1,
                          bbox_.iymin, bbox_.iymax - 1)
                         for bbox_ in bbox])

    @lazyproperty
    @as_scalar
    def bbox_xmin(self):
        """
        The minimum ``x``-pixel index of the bounding box.
        """
        return np.transpose(self._bbox_bounds)[0]

    @lazyproperty
    @as_scalar
    def bbox_xmax(self):
        """
        The maximum ``x``-pixel index of the bounding box.

        Note that this value is inclusive, unlike numpy slice indices.
        """
        return np.transpose(self._bbox_bounds)[1]

    @lazyproperty
    @as_scalar
    def bbox_ymin(self):
        """
        The minimum ``y``-pixel index of the bounding box.
        """
        return np.transpose(self._bbox_bounds)[2]

    @lazyproperty
    @as_scalar
    def bbox_ymax(self):
        """
        The maximum ``y``-pixel index of the bounding box.

        Note that this value is inclusive, unlike numpy slice indices.
        """
        return np.transpose(self._bbox_bounds)[3]

    def _calculate_stats(self, stat_func, *, unit=None):
        """
        Apply the input ``stat_func`` to the 1D array of unmasked data
        values in the aperture.

        Units are applied if the input ``data`` has units.

        Parameters
        ----------
        stat_func : callable
            The callable to apply to the 1D `~numpy.ndarray` of unmasked
            data values.

        unit : `None` or `astropy.unit.Unit`, optional
            The unit to apply to the output data. This is used only
            if the input ``data`` has units. If `None` then the input
            ``data`` unit will be used.
        """
        result = np.array([stat_func(arr) for arr in self._data_values_center])
        if unit is None:
            unit = self._data_unit
        if unit is not None:
            result <<= unit
        return result

    @lazyproperty
    @as_scalar
    def center_aper_area(self):
        """
        The total area of the unmasked pixels within the aperture using
        the "center" aperture mask method.
        """
        areas = np.array([np.sum(weight.filled(0.0))
                          for weight in self._weight_cutout_center])
        areas[self._all_masked] = np.nan
        return areas << (u.pix**2)

    @lazyproperty
    @as_scalar
    def sum_aper_area(self):
        """
        The total area of the unmasked pixels within the aperture using
        the input ``sum_method`` aperture mask method.
        """
        areas = np.array([np.sum(weight.filled(0.0))
                          for weight in self._weight_cutout])
        areas[self._all_masked] = np.nan
        return areas << (u.pix**2)

    @lazyproperty
    @as_scalar
    def sum(self):
        r"""
        The sum of the unmasked ``data`` values within the aperture.

        .. math::

            F = \sum_{i \in A} I_i

        where :math:`F` is ``sum``, :math:`I_i` is the
        background-subtracted ``data``, and :math:`A` are the unmasked
        pixels in the aperture.

        Non-finite pixel values (NaN and inf) are excluded
        (automatically masked).
        """
        if self.sum_method == 'center':
            return self._calculate_stats(np.sum)

        data_values = self._get_values(self.data_sum_cutout)
        result = np.array([np.sum(arr) for arr in data_values])
        if self._data_unit is not None:
            result <<= self._data_unit
        return result

    @lazyproperty
    @as_scalar
    def sum_err(self):
        r"""
        The uncertainty of `sum`, propagated from the input ``error``
        array.

        ``sum_err`` is the quadrature sum of the total errors over the
        unmasked pixels within the aperture:

        .. math::

            \Delta F = \sqrt{\sum_{i \in A} \sigma_{\mathrm{tot}, i}^2}

        where :math:`\Delta F` is the `sum_err`,
        :math:`\sigma_{\mathrm{tot, i}}` are the pixel-wise total errors
        (``error``), and :math:`A` are the unmasked pixels in the
        aperture.

        Pixel values that are masked in the input ``data``, including
        any non-finite pixel values (NaN and inf) that are automatically
        masked, are also masked in the error array.
        """
        if self._error is None:
            err = self._null_value

        else:
            if self.sum_method == 'center':
                variance = self._variance_cutout_center
            else:
                variance = self._variance_cutout

            var_values = [arr.compressed() if len(arr.compressed()) > 0
                          else np.array([np.nan]) for arr in variance]
            err = np.sqrt([np.sum(arr) for arr in var_values])

        if self._data_unit is not None:
            err <<= self._data_unit
        return err

    @lazyproperty
    @as_scalar
    def min(self):
        """
        The minimum of the unmasked pixel values within the aperture.
        """
        return self._calculate_stats(np.min)

    @lazyproperty
    @as_scalar
    def max(self):
        """
        The maximum of the unmasked pixel values within the aperture.
        """
        return self._calculate_stats(np.max)

    @lazyproperty
    @as_scalar
    def mean(self):
        """
        The mean of the unmasked pixel values within the aperture.
        """
        return self._calculate_stats(np.mean)

    @lazyproperty
    @as_scalar
    def median(self):
        """
        The median of the unmasked pixel values within the aperture.
        """
        return self._calculate_stats(np.median)

    @lazyproperty
    @as_scalar
    def mode(self):
        """
        The mode of the unmasked pixel values within the aperture.

        The mode is estimated as ``(3 * median) - (2 * mean)``.
        """
        return 3.0 * self.median - 2.0 * self.mean

    @lazyproperty
    @as_scalar
    def std(self):
        """
        The standard deviation of the unmasked pixel values within the
        aperture.
        """
        return self._calculate_stats(np.std)

    @lazyproperty
    @as_scalar
    def mad_std(self):
        r"""
        The standard deviation calculated using
        the `median absolute deviation (MAD)
        <https://en.wikipedia.org/wiki/Median_absolute_deviation>`_.

        The standard deviation estimator is given by:

        .. math::

            \sigma \approx \frac{\textrm{MAD}}{\Phi^{-1}(3/4)}
                \approx 1.4826 \ \textrm{MAD}

        where :math:`\Phi^{-1}(P)` is the normal inverse cumulative
        distribution function evaluated at probability :math:`P = 3/4`.
        """
        return self._calculate_stats(mad_std)

    @lazyproperty
    @as_scalar
    def var(self):
        """
        The variance of the unmasked pixel values within the aperture.
        """
        unit = self._data_unit
        if unit is not None:
            unit **= 2
        return self._calculate_stats(np.var, unit=unit)

    @lazyproperty
    @as_scalar
    def biweight_location(self):
        """
        The biweight location of the unmasked pixel values within the
        aperture.

        See `astropy.stats.biweight_location`.
        """
        return self._calculate_stats(biweight_location)

    @lazyproperty
    @as_scalar
    def biweight_midvariance(self):
        """
        The biweight midvariance of the unmasked pixel values within the
        aperture.

        See `astropy.stats.biweight_midvariance`
        """
        unit = self._data_unit
        if unit is not None:
            unit **= 2
        return self._calculate_stats(biweight_midvariance, unit=unit)

    @lazyproperty
    @as_scalar
    def inertia_tensor(self):
        """
        The inertia tensor of the source for the rotation around its
        center of mass.
        """
        moments = self.moments_central
        if self.isscalar:
            moments = moments[np.newaxis, :]
        mu_02 = moments[:, 0, 2]
        mu_11 = -moments[:, 1, 1]
        mu_20 = moments[:, 2, 0]
        tensor = np.array([mu_02, mu_11, mu_11, mu_20]).swapaxes(0, 1)
        return tensor.reshape((tensor.shape[0], 2, 2)) * u.pix**2

    @lazyproperty
    def _covariance(self):
        """
        The covariance matrix of the 2D Gaussian function that has the
        same second-order moments as the source, always as an iterable.
        """
        moments = self.moments_central
        if self.isscalar:
            moments = moments[np.newaxis, :]
        # Ignore divide-by-zero RuntimeWarning
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            mu_norm = moments / moments[:, 0, 0][:, np.newaxis, np.newaxis]

        covar = np.array([mu_norm[:, 0, 2], mu_norm[:, 1, 1],
                          mu_norm[:, 1, 1], mu_norm[:, 2, 0]]).swapaxes(0, 1)
        covar = covar.reshape((covar.shape[0], 2, 2))

        # Modify the covariance matrix in the case of "infinitely" thin
        # detections. This follows SourceExtractor's prescription of
        # incrementally increasing the diagonal elements by 1/12.
        delta = 1.0 / 12
        delta2 = delta**2
        # Ignore RuntimeWarning from NaN values in covar
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            covar_det = np.linalg.det(covar)

            # Covariance should be positive semidefinite
            idx = np.where(covar_det < 0)[0]
            covar[idx] = np.array([[np.nan, np.nan], [np.nan, np.nan]])

            idx = np.where(covar_det < delta2)[0]
            while idx.size > 0:
                covar[idx, 0, 0] += delta
                covar[idx, 1, 1] += delta
                covar_det = np.linalg.det(covar)
                idx = np.where(covar_det < delta2)[0]
        return covar

    @lazyproperty
    @as_scalar
    def covariance(self):
        """
        The covariance matrix of the 2D Gaussian function that has the
        same second-order moments as the source.
        """
        return self._covariance * (u.pix**2)

    @lazyproperty
    @as_scalar
    def covariance_eigvals(self):
        """
        The two eigenvalues of the `covariance` matrix in decreasing
        order.
        """
        eigvals = np.empty((self.n_apertures, 2))
        eigvals.fill(np.nan)
        # np.linalg.eigvalsh requires finite input values
        idx = np.unique(np.where(np.isfinite(self._covariance))[0])
        eigvals[idx] = np.linalg.eigvalsh(self._covariance[idx])

        # Check for negative variance
        # (just in case covariance matrix is not positive semidefinite)
        idx2 = np.unique(np.where(eigvals < 0)[0])
        eigvals[idx2] = (np.nan, np.nan)

        # Sort each eigenvalue pair in descending order
        # (eigvalsh returns values in ascending order)
        eigvals = np.fliplr(eigvals)

        return eigvals * u.pix**2

    @lazyproperty
    @as_scalar
    def semimajor_axis(self):
        """
        The 1-sigma standard deviation along the semimajor axis of the
        2D Gaussian function that has the same second-order central
        moments as the source.
        """
        eigvals = self.covariance_eigvals
        if self.isscalar:
            eigvals = eigvals[np.newaxis, :]
        return np.sqrt(eigvals[:, 0])

    @lazyproperty
    @as_scalar
    def semiminor_axis(self):
        """
        The 1-sigma standard deviation along the semiminor axis of the
        2D Gaussian function that has the same second-order central
        moments as the source.
        """
        eigvals = self.covariance_eigvals
        if self.isscalar:
            eigvals = eigvals[np.newaxis, :]
        return np.sqrt(eigvals[:, 1])

    @lazyproperty
    @as_scalar
    def fwhm(self):
        r"""
        The circularized full width at half maximum (FWHM) of the 2D
        Gaussian function that has the same second-order central moments
        as the source.

        .. math::

           \mathrm{FWHM} & = 2 \sqrt{2 \ln(2)} \sqrt{0.5 (a^2 + b^2)}
           \\
                         & = 2 \sqrt{\ln(2) \ (a^2 + b^2)}

        where :math:`a` and :math:`b` are the 1-sigma lengths of the
        semimajor (`semimajor_axis`) and semiminor (`semiminor_axis`)
        axes, respectively.
        """
        return 2.0 * np.sqrt(np.log(2.0) * (self.semimajor_axis**2
                                            + self.semiminor_axis**2))

    @lazyproperty
    @as_scalar
    def orientation(self):
        """
        The angle between the ``x`` axis and the major axis of the 2D
        Gaussian function that has the same second-order moments as the
        source.

        The angle increases in the counter-clockwise direction.
        """
        covar = self._covariance
        orient_radians = 0.5 * np.arctan2(2.0 * covar[:, 0, 1],
                                          (covar[:, 0, 0] - covar[:, 1, 1]))
        return np.rad2deg(orient_radians) * u.deg

    @lazyproperty
    @as_scalar
    def eccentricity(self):
        r"""
        The eccentricity of the 2D Gaussian function that has the same
        second-order moments as the source.

        The eccentricity is the fraction of the distance along the
        semimajor axis at which the focus lies.

        .. math::

            e = \sqrt{1 - \frac{b^2}{a^2}}

        where :math:`a` and :math:`b` are the lengths of the semimajor
        and semiminor axes, respectively.
        """
        semimajor_var, semiminor_var = np.transpose(self.covariance_eigvals)
        return np.sqrt(1.0 - (semiminor_var / semimajor_var))

    @lazyproperty
    @as_scalar
    def elongation(self):
        r"""
        The ratio of the lengths of the semimajor and semiminor axes.

        .. math::

            \mathrm{elongation} = \frac{a}{b}

        where :math:`a` and :math:`b` are the lengths of the semimajor
        and semiminor axes, respectively.
        """
        return self.semimajor_axis / self.semiminor_axis

    @lazyproperty
    @as_scalar
    def ellipticity(self):
        r"""
        1.0 minus the ratio of the lengths of the semimajor and
        semiminor axes (or 1.0 minus the `elongation`).

        .. math::

            \mathrm{ellipticity} = 1 - \frac{b}{a}

        where :math:`a` and :math:`b` are the lengths of the semimajor
        and semiminor axes, respectively.
        """
        return 1.0 - (self.semiminor_axis / self.semimajor_axis)

    @lazyproperty
    @as_scalar
    def covariance_xx(self):
        r"""
        The ``(0, 0)`` element of the `covariance` matrix, representing
        :math:`\sigma_x^2`, in units of pixel**2.
        """
        return self._covariance[:, 0, 0] * u.pix**2

    @lazyproperty
    @as_scalar
    def covariance_yy(self):
        r"""
        The ``(1, 1)`` element of the `covariance` matrix, representing
        :math:`\sigma_y^2`, in units of pixel**2.
        """
        return self._covariance[:, 1, 1] * u.pix**2

    @lazyproperty
    @as_scalar
    def covariance_xy(self):
        r"""
        The ``(0, 1)`` and ``(1, 0)`` elements of the `covariance`
        matrix, representing :math:`\sigma_x \sigma_y`, in units of
        pixel**2.
        """
        return self._covariance[:, 0, 1] * u.pix**2

    @lazyproperty
    @as_scalar
    def ellipse_cxx(self):
        r"""
        Coefficient for ``x**2`` in the generalized ellipse equation in
        units of pixel**(-2).

        The ellipse is defined as

        .. math::

            cxx (x - \bar{x})^2 + cxy (x - \bar{x}) (y - \bar{y}) +
            cyy (y - \bar{y})^2 = R^2

        where :math:`R` is a parameter which scales the ellipse (in
        units of the axes lengths).

        `SourceExtractor`_ reports that the isophotal limit of a source
        is well represented by :math:`R \approx 3`.
        """
        return ((np.cos(self.orientation) / self.semimajor_axis)**2
                + (np.sin(self.orientation) / self.semiminor_axis)**2)

    @lazyproperty
    @as_scalar
    def ellipse_cyy(self):
        r"""
        Coefficient for ``y**2`` in the generalized ellipse equation in
        units of pixel**(-2).

        The ellipse is defined as

        .. math::

            cxx (x - \bar{x})^2 + cxy (x - \bar{x}) (y - \bar{y}) +
            cyy (y - \bar{y})^2 = R^2

        where :math:`R` is a parameter which scales the ellipse (in
        units of the axes lengths).

        `SourceExtractor`_ reports that the isophotal limit of a source
        is well represented by :math:`R \approx 3`.
        """
        return ((np.sin(self.orientation) / self.semimajor_axis)**2
                + (np.cos(self.orientation) / self.semiminor_axis)**2)

    @lazyproperty
    @as_scalar
    def ellipse_cxy(self):
        r"""
        Coefficient for ``x * y`` in the generalized ellipse equation in
        units of pixel**(-2).

        The ellipse is defined as

        .. math::

            cxx (x - \bar{x})^2 + cxy (x - \bar{x}) (y - \bar{y}) +
            cyy (y - \bar{y})^2 = R^2

        where :math:`R` is a parameter which scales the ellipse (in
        units of the axes lengths).

        `SourceExtractor`_ reports that the isophotal limit of a source
        is well represented by :math:`R \approx 3`.
        """
        return (2.0 * np.cos(self.orientation) * np.sin(self.orientation)
                * ((1.0 / self.semimajor_axis**2)
                   - (1.0 / self.semiminor_axis**2)))

    @lazyproperty
    @as_scalar
    def gini(self):
        r"""
        The `Gini coefficient
        <https://en.wikipedia.org/wiki/Gini_coefficient>`_ of the
        unmasked pixel values within the aperture.

        The Gini coefficient of the distribution of absolute flux values
        is calculated using the prescription from `Lotz et al. 2004
        <https://ui.adsabs.harvard.edu/abs/2004AJ....128..163L/abstract>`_
        (Eq. 6) as:

        .. math::

            G = \frac{1}{\overline{|x|} \, n \, (n - 1)}
                \sum^{n}_{i} (2i - n - 1) \left | x_i \right |

        where :math:`\overline{|x|}` is the mean of the absolute value
        of all pixel values :math:`x_i`. If the sum of all pixel values
        is zero, the Gini coefficient is zero.

        Negative pixel values are used via their absolute value. Invalid
        values (NaN and inf) in the input are automatically excluded
        from the calculation. If only a single finite pixel remains
        after filtering, the Gini coefficient is 0.0.
        """
        return np.array([gini_func(arr) for arr in self._data_values_center])



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the functional_models module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.modeling.fitting import TRFLSQFitter
from astropy.stats import gaussian_fwhm_to_sigma
from numpy.testing import assert_allclose

from photutils.psf import (AiryDiskPSF, CircularGaussianPRF,
                           CircularGaussianPSF, CircularGaussianSigmaPRF,
                           GaussianPRF, GaussianPSF, MoffatPSF)


def make_gaussian_models(name):
    flux = 71.4
    x_0 = 24.3
    y_0 = 25.2
    x_fwhm = 10.1
    y_fwhm = 5.82
    theta = 21.7
    flux_i = 50
    x_0_i = 20
    y_0_i = 30
    x_fwhm_i = 15
    y_fwhm_i = 8
    theta_i = 31
    if name == 'GaussianPSF':
        model = GaussianPSF(flux=flux, x_0=x_0, y_0=y_0, x_fwhm=x_fwhm,
                            y_fwhm=y_fwhm, theta=theta)
        model_init = GaussianPSF(flux=flux_i, x_0=x_0_i, y_0=y_0_i,
                                 x_fwhm=x_fwhm_i, y_fwhm=y_fwhm_i,
                                 theta=theta_i)
    elif name == 'GaussianPRF':
        model = GaussianPRF(flux=flux, x_0=x_0, y_0=y_0, x_fwhm=x_fwhm,
                            y_fwhm=y_fwhm, theta=theta)
        model_init = GaussianPRF(flux=flux_i, x_0=x_0_i, y_0=y_0_i,
                                 x_fwhm=x_fwhm_i, y_fwhm=y_fwhm_i,
                                 theta=theta_i)
    elif name == 'CircularGaussianPSF':
        model = CircularGaussianPSF(flux=flux, x_0=x_0, y_0=y_0, fwhm=x_fwhm)
        model_init = CircularGaussianPSF(flux=flux_i, x_0=x_0_i, y_0=y_0_i,
                                         fwhm=x_fwhm_i)
    elif name == 'CircularGaussianPRF':
        model = CircularGaussianPRF(flux=flux, x_0=x_0, y_0=y_0, fwhm=x_fwhm)
        model_init = CircularGaussianPRF(flux=flux_i, x_0=x_0_i, y_0=y_0_i,
                                         fwhm=x_fwhm_i)
    elif name == 'CircularGaussianSigmaPRF':
        model = CircularGaussianSigmaPRF(flux=flux, x_0=x_0, y_0=y_0,
                                         sigma=x_fwhm / 2.35)
        model_init = CircularGaussianSigmaPRF(flux=flux_i, x_0=x_0_i,
                                              y_0=y_0_i, sigma=x_fwhm_i / 2.35)
    else:
        msg = 'invalid model name'
        raise ValueError(msg)

    return model, model_init


def gaussian_tests(name, use_units):
    model, model_init = make_gaussian_models(name)
    fixed_types = ('fwhm', 'sigma', 'theta')
    for param in model.param_names:
        for fixed_type in fixed_types:
            if fixed_type in param:
                tparam = getattr(model_init, param)
                tparam.fixed = False

    yy, xx = np.mgrid[0:51, 0:51]

    if use_units:
        unit = u.m
        xx <<= unit
        yy <<= unit

        unit_params = ('x_0', 'y_0', 'x_fwhm', 'y_fwhm', 'fwhm', 'sigma')
        for param in model.param_names:
            if param in unit_params:
                tparam = getattr(model, param)
                tparam <<= unit
                setattr(model, param, tparam)

    data = model(xx, yy)
    if use_units:
        data = data.value
    assert_allclose(data.sum(), model.flux.value)
    try:
        assert_allclose(model.x_sigma, model.x_fwhm * gaussian_fwhm_to_sigma)
        assert_allclose(model.y_sigma, model.y_fwhm * gaussian_fwhm_to_sigma)
    except AttributeError:
        assert_allclose(model.sigma, model.fwhm * gaussian_fwhm_to_sigma)

    try:
        xsigma = model.x_sigma
        ysigma = model.y_sigma
        if isinstance(xsigma, u.Quantity):
            xsigma = xsigma.value
            ysigma = ysigma.value
        assert_allclose(model.amplitude * (2 * np.pi * xsigma * ysigma),
                        model.flux)
    except AttributeError:
        sigma = model.sigma
        if isinstance(sigma, u.Quantity):
            sigma = sigma.value
        assert_allclose(model.amplitude * (2 * np.pi * sigma**2), model.flux)

    fitter = TRFLSQFitter()
    fit_model = fitter(model_init, xx, yy, data)
    assert_allclose(fit_model.x_0.value, model.x_0.value, rtol=1e-5)
    assert_allclose(fit_model.y_0.value, model.y_0.value, rtol=1e-5)
    try:
        assert_allclose(fit_model.x_fwhm.value, model.x_fwhm.value)
        assert_allclose(fit_model.y_fwhm.value, model.y_fwhm.value)
        assert_allclose(fit_model.theta.value, model.theta.value)
    except AttributeError:
        if name == 'CircularGaussianSigmaPRF':
            assert_allclose(fit_model.sigma.value, model.sigma.value)
        else:
            assert_allclose(fit_model.fwhm.value, model.fwhm.value)

    # test the model derivatives
    fit_model2 = fitter(model_init, xx, yy, data, estimate_jacobian=True)
    assert_allclose(fit_model2.x_0, fit_model.x_0)
    assert_allclose(fit_model2.y_0, fit_model.y_0)
    try:
        assert_allclose(fit_model2.x_fwhm, fit_model.x_fwhm)
        assert_allclose(fit_model2.y_fwhm, fit_model.y_fwhm)
        assert_allclose(fit_model2.theta, fit_model.theta)
    except AttributeError:
        assert_allclose(fit_model2.fwhm, fit_model.fwhm)

    if use_units and 'Circular' not in name:
        model.y_0 = model.y_0.value * u.s
        yy = yy.value * u.s
        match = 'Units .* inputs should match'
        with pytest.raises(u.UnitsError, match=match):
            fitter(model_init, xx, yy, data)


@pytest.mark.parametrize('name', ['GaussianPSF', 'CircularGaussianPSF'])
@pytest.mark.parametrize('use_units', [False, True])
def test_gaussian_psfs(name, use_units):
    gaussian_tests(name, use_units)


@pytest.mark.parametrize('name', ['GaussianPRF', 'CircularGaussianPRF',
                                  'CircularGaussianSigmaPRF'])
@pytest.mark.parametrize('use_units', [False, True])
def test_gaussian_prfs(name, use_units):
    gaussian_tests(name, use_units)


def test_gaussian_prf_sums():
    """
    Test that subpixel accuracy of Gaussian PRFs by checking the sum of
    pixels.
    """
    model1 = GaussianPRF(x_0=0, y_0=0, x_fwhm=0.001, y_fwhm=0.001)
    model2 = CircularGaussianPRF(x_0=0, y_0=0, fwhm=0.001)
    model3 = CircularGaussianSigmaPRF(x_0=0, y_0=0, sigma=0.001)
    yy, xx = np.mgrid[-10:11, -10:11]
    for model in (model1, model2, model3):
        assert_allclose(model(xx, yy).sum(), 1.0)


def test_gaussian_bounding_boxes():
    model1 = GaussianPSF(x_0=0, y_0=0, x_fwhm=2, y_fwhm=3)
    model2 = GaussianPRF(x_0=0, y_0=0, x_fwhm=2, y_fwhm=3)
    xbbox = (-4.6712699, 4.6712699)
    ybbox = (-7.0069049, 7.0069048)
    for model in (model1, model2):
        assert_allclose(model.bounding_box, (xbbox, ybbox))

    model3 = CircularGaussianPSF(x_0=0, y_0=0, fwhm=2)
    model4 = CircularGaussianPRF(x_0=0, y_0=0, fwhm=2)
    for model in (model3, model4):
        assert_allclose(model.bounding_box, (xbbox, xbbox))

    model5 = CircularGaussianSigmaPRF(x_0=0, y_0=0, sigma=2)
    assert_allclose(model5.bounding_box, ((-11, 11), (-11, 11)))


@pytest.mark.parametrize('use_units', [False, True])
def test_moffat_psf_model(use_units):
    model = MoffatPSF(flux=71.4, x_0=24.3, y_0=25.2, alpha=8.1, beta=7.2)
    model_init = MoffatPSF(flux=50, x_0=20, y_0=30, alpha=5, beta=4)
    model_init.alpha.fixed = False
    model_init.beta.fixed = False

    yy, xx = np.mgrid[0:51, 0:51]

    if use_units:
        unit = u.cm
        xx <<= unit
        yy <<= unit
        model.x_0 <<= unit
        model.y_0 <<= unit
        model.alpha <<= unit

    data = model(xx, yy)
    assert_allclose(data.sum(), model.flux.value, rtol=5e-6)
    fwhm = 2 * model.alpha * np.sqrt(2**(1 / model.beta) - 1)
    assert_allclose(model.fwhm, fwhm)

    fitter = TRFLSQFitter()
    fit_model = fitter(model_init, xx, yy, data)
    assert_allclose(fit_model.x_0.value, model.x_0.value)
    assert_allclose(fit_model.y_0.value, model.y_0.value)
    assert_allclose(fit_model.alpha.value, model.alpha.value)
    assert_allclose(fit_model.beta.value, model.beta.value)

    # test bounding box
    model = MoffatPSF(x_0=0, y_0=0, alpha=1.0, beta=2.0)
    bbox = 12.871885058111655
    assert_allclose(model.bounding_box, ((-bbox, bbox), (-bbox, bbox)))


@pytest.mark.parametrize('use_units', [False, True])
def test_airydisk_psf_model(use_units):
    model = AiryDiskPSF(flux=71.4, x_0=24.3, y_0=25.2, radius=2.1)
    model_init = AiryDiskPSF(flux=50, x_0=23, y_0=27, radius=2.5)
    model_init.radius.fixed = False

    yy, xx = np.mgrid[0:51, 0:51]

    if use_units:
        unit = u.cm
        xx <<= unit
        yy <<= unit
        model.x_0 <<= unit
        model.y_0 <<= unit
        model.radius <<= unit

    data = model(xx, yy)
    assert_allclose(data.sum(), model.flux.value, rtol=0.015)
    fwhm = 0.8436659602162364 * model.radius
    assert_allclose(model.fwhm, fwhm)

    fitter = TRFLSQFitter()
    fit_model = fitter(model_init, xx, yy, data)
    assert_allclose(fit_model.x_0.value, model.x_0.value)
    assert_allclose(fit_model.y_0.value, model.y_0.value)
    assert_allclose(fit_model.radius.value, model.radius.value)

    # test bounding box
    model = AiryDiskPSF(x_0=0, y_0=0, radius=5)
    bbox = 42.18329801081182
    assert_allclose(model.bounding_box, ((-bbox, bbox), (-bbox, bbox)))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for Quantity helpers.
"""

import astropy.units as u
import numpy as np


def check_units(values, names):
    """
    Check that input values have consistent units.

    Parameters
    ----------
    values : list of scalar, `~numpy.ndarray`, or `~astropy.units.Quantity`
        A list of values.

    names : list of str
        A list of names corresponding to the input ``values``.

    Returns
    -------
    units : set
        The set of distinct units across all non-`None` values.

    Raises
    ------
    ValueError
        If the number of values does not match the number of names,
        or if the input values do not all have the same units.
    """
    if len(values) != len(names):
        msg = 'The number of values must match the number of names.'
        raise ValueError(msg)

    all_units = {name: getattr(arr, 'unit', None)
                 for arr, name in zip(values, names, strict=True)
                 if arr is not None}
    units = set(all_units.values())

    if len(units) > 1:
        param_names = list(all_units.keys())
        msg = [f'The inputs {param_names} must all have the same units:']
        indent = ' ' * 4
        for key, value in all_units.items():
            if value is None:
                msg.append(f'{indent}{key} does not have units')
            else:
                msg.append(f'{indent}{key} has units of {value}')
        msg = '\n'.join(msg)
        raise ValueError(msg)

    return units


def process_quantities(values, names):
    """
    Check and remove units of input values.

    If any of the input values have units then they all must have
    units and the units must be the same.

    The returned values are the input values with units removed and the
    unit.

    Parameters
    ----------
    values : list of scalar, `~numpy.ndarray`, or `~astropy.units.Quantity`
        A list of values.

    names : list of str
        A list of names corresponding to the input ``values``.

    Returns
    -------
    values : list of scalar or `~numpy.ndarray`
        A list of values, where units have been removed.

    unit : `~astropy.unit.Unit`
        The common unit for the input values. `None` will be returned
        if all the input values do not have units (including when all
        values are `None`).

    Raises
    ------
    ValueError
        If the input values do not all have the same units.
    """
    units = check_units(values, names)

    # When all values are None, the units set is empty; return unchanged
    # with unit=None
    if len(units) == 0:
        return values, None

    # Extract the unit and remove it from the return values
    unit = units.pop()
    if unit is not None:
        values = [val.value if val is not None else val for val in values]

    return values, unit


def isscalar(value):
    """
    Check if a value is a scalar.

    This works for both `~astropy.units.Quantity` and scalars.

    `numpy.isscalar` always returns False for `~astropy.units.Quantity`
    objects.

    Parameters
    ----------
    value : `~astropy.units.Quantity`, scalar, or array_like
        The value to check.

    Returns
    -------
    isscalar : bool
        `True` if the value is a scalar, `False` otherwise.
    """
    if isinstance(value, u.Quantity):
        return value.isscalar

    return np.isscalar(value)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the load module.
"""

from unittest.mock import patch
from urllib.error import URLError

import numpy as np
import pytest
from astropy.io import fits
from astropy.table import Table
from astropy.utils.exceptions import AstropyDeprecationWarning

from photutils.datasets import load
from photutils.datasets.load import _get_path, _load_fits_as_imagehdu


def test_get_path():
    """
    Test _get_path with a valid filename and location, and with an
    invalid location.
    """
    fn = '4gaussians_params.ecsv'
    path = _get_path(fn, location='local')
    assert fn in path

    match = 'Invalid location:'
    with pytest.raises(ValueError, match=match):
        _get_path('filename', location='invalid')


def test_get_path_photutils_datasets():
    """
    Test _get_path with location='photutils-datasets'.
    """
    with patch('photutils.datasets.load.download_file') as mock_dl:
        mock_dl.return_value = '/path/to/file.fits'
        result = _get_path('file.fits', location='photutils-datasets',
                           cache=False)
        assert result == '/path/to/file.fits'
        mock_dl.assert_called_once()
        call_args = mock_dl.call_args
        assert 'photutils-datasets' in call_args[0][0]
        assert call_args[1]['cache'] is False


@pytest.fixture
def url_paths():
    """
    Fixture providing URLs for cache tests.
    """
    filename = 'test_file.fits'
    primary_url = f'http://data.astropy.org/photometry/{filename}'
    datasets_url = (
        'https://github.com/astropy/photutils-datasets/raw/'
        f'main/data/{filename}'
    )
    return {
        'filename': filename,
        'primary_url': primary_url,
        'datasets_url': datasets_url,
    }


class TestGetPathCache:
    """
    Tests for the caching behavior of _get_path.

    Tests that it correctly checks the cache for both the primary and
    fallback URLs, and that it falls back to downloading from the
    datasets URL if the primary URL is not cached and fails to download.
    """

    def test_cache_hit_primary_url(self, url_paths):
        """
        Test that _get_path uses the cached file when the primary URL is
        already in the cache, without trying the fallback URL.
        """
        with (
            patch('photutils.datasets.load.is_url_in_cache') as mock_cache,
            patch('photutils.datasets.load.download_file') as mock_dl,
        ):
            mock_cache.side_effect = (
                lambda url: url == url_paths['primary_url']
            )
            mock_dl.return_value = '/cached/path/test_file.fits'

            result = _get_path(url_paths['filename'], location='remote')

            assert result == '/cached/path/test_file.fits'
            mock_dl.assert_called_once_with(
                url_paths['primary_url'], cache=True, show_progress=False,
            )

    def test_cache_hit_datasets_url(self, url_paths):
        """
        Test that _get_path uses the cached file when only the fallback
        datasets URL is in the cache.
        """
        with (
            patch('photutils.datasets.load.is_url_in_cache') as mock_cache,
            patch('photutils.datasets.load.download_file') as mock_dl,
        ):
            mock_cache.side_effect = (
                lambda url: url == url_paths['datasets_url']
            )
            mock_dl.return_value = '/cached/path/test_file.fits'

            result = _get_path(url_paths['filename'], location='remote')

            assert result == '/cached/path/test_file.fits'
            mock_dl.assert_called_once_with(
                url_paths['datasets_url'], cache=True, show_progress=False,
            )

    def test_no_cache_falls_through_to_download(self, url_paths):
        """
        Test that _get_path tries the primary URL and falls back to the
        datasets URL when neither is cached and the primary fails.
        """
        with (
            patch('photutils.datasets.load.is_url_in_cache',
                  return_value=False),
            patch('photutils.datasets.load.download_file') as mock_dl,
        ):
            mock_dl.side_effect = [
                URLError('timeout'),
                '/downloaded/path/test_file.fits',
            ]

            result = _get_path(url_paths['filename'], location='remote')

            assert result == '/downloaded/path/test_file.fits'
            assert mock_dl.call_count == 2


def test_load_irac_psf_invalid_channel():
    """
    Test that load_irac_psf raises a ValueError when an invalid channel
    number is provided.
    """
    match = 'channel must be 1, 2, 3, or 4'
    with pytest.raises(ValueError, match=match):
        load.load_irac_psf(0)
    with pytest.raises(ValueError, match=match):
        load.load_irac_psf(5)


@pytest.mark.remote_data
def test_load_star_image():
    """
    Test that load_star_image returns an HDU with the expected header
    and data shape.
    """
    with pytest.warns(AstropyDeprecationWarning):
        hdu = load.load_star_image()
    assert len(hdu.header) == 106
    assert hdu.data.shape == (1059, 1059)


def test_load_fits_as_imagehdu(tmp_path):
    """
    Test the _load_fits_as_imagehdu helper function.
    """
    data = np.ones((10, 10))
    header = fits.Header()
    header['SIMPLE'] = True
    header['BITPIX'] = -32
    header['NAXIS'] = 2
    header['NAXIS1'] = 10
    header['NAXIS2'] = 10
    header['EXTEND'] = True
    header['COMMENT'] = 'Test header'

    primary_hdu = fits.PrimaryHDU(data=data, header=header)
    hdulist = fits.HDUList([primary_hdu])

    filepath = tmp_path / 'test.fits'
    hdulist.writeto(filepath)

    result = _load_fits_as_imagehdu(str(filepath))

    assert isinstance(result, fits.ImageHDU)
    assert np.array_equal(result.data, data)
    assert result.header['COMMENT'] == 'Test header'


class TestLoadFunctionsMocked:
    """
    Tests for the load functions using mocking to avoid remote data
    access.
    """

    def test_load_spitzer_image(self, tmp_path):
        """
        Test the load_spitzer_image function with mocked file download.
        """
        data = np.random.default_rng(seed=0).random((100, 100))
        header = fits.Header()
        header['TELESCOP'] = 'Spitzer'
        header['INSTRUME'] = 'IRAC'

        primary_hdu = fits.PrimaryHDU(data=data, header=header)
        hdulist = fits.HDUList([primary_hdu])

        filepath = tmp_path / 'spitzer_example_image.fits'
        hdulist.writeto(filepath)

        with (pytest.warns(AstropyDeprecationWarning),
              patch('photutils.datasets.load._get_path',
                    return_value=str(filepath))):
            hdu = load.load_spitzer_image()

        assert isinstance(hdu, fits.ImageHDU)
        assert np.array_equal(hdu.data, data)
        assert hdu.header['TELESCOP'] == 'Spitzer'

    def test_load_spitzer_image_show_progress(self, tmp_path):
        """
        Test the load_spitzer_image function with show_progress=True.
        """
        data = np.random.default_rng(seed=0).random((100, 100))
        header = fits.Header()

        primary_hdu = fits.PrimaryHDU(data=data, header=header)
        hdulist = fits.HDUList([primary_hdu])

        filepath = tmp_path / 'spitzer_example_image.fits'
        hdulist.writeto(filepath)

        with (pytest.warns(AstropyDeprecationWarning),
              patch('photutils.datasets.load._get_path',
                    return_value=str(filepath)) as mock_get_path):
            hdu = load.load_spitzer_image(show_progress=True)

        assert isinstance(hdu, fits.ImageHDU)
        mock_get_path.assert_called_once_with(
            'spitzer_example_image.fits',
            location='remote',
            show_progress=True,
        )

    def test_load_spitzer_catalog(self, tmp_path):
        """
        Test the load_spitzer_catalog function with mocked file
        download.
        """
        catalog_data = Table()
        catalog_data['l'] = [18.23, 18.15, 18.30]
        catalog_data['b'] = [0.20, 0.22, 0.18]

        filepath = tmp_path / 'spitzer_example_catalog.xml'
        catalog_data.write(filepath, format='votable', overwrite=True)

        with (pytest.warns(AstropyDeprecationWarning),
              patch('photutils.datasets.load._get_path',
                    return_value=str(filepath))):
            catalog = load.load_spitzer_catalog()

        assert isinstance(catalog, Table)
        assert 'l' in catalog.colnames
        assert 'b' in catalog.colnames
        assert len(catalog) == 3

    def test_load_spitzer_catalog_show_progress(self, tmp_path):
        """
        Test the load_spitzer_catalog function with show_progress=True.
        """
        catalog_data = Table()
        catalog_data['l'] = [18.23]
        catalog_data['b'] = [0.20]

        filepath = tmp_path / 'spitzer_example_catalog.xml'
        catalog_data.write(filepath, format='votable', overwrite=True)

        with (pytest.warns(AstropyDeprecationWarning),
              patch('photutils.datasets.load._get_path',
                    return_value=str(filepath)) as mock_get_path):
            catalog = load.load_spitzer_catalog(show_progress=True)

        assert isinstance(catalog, Table)
        mock_get_path.assert_called_once_with(
            'spitzer_example_catalog.xml',
            location='remote',
            show_progress=True,
        )

    def test_load_irac_psf(self, tmp_path):
        """
        Test the load_irac_psf function with mocked file download.
        """
        data = np.random.default_rng(seed=0).random((41, 41))
        header = fits.Header()
        header['TELESCOP'] = 'Spitzer'
        header['INSTRUME'] = 'IRAC'
        header['CHNLNUM'] = 1

        primary_hdu = fits.PrimaryHDU(data=data, header=header)
        hdulist = fits.HDUList([primary_hdu])

        filepath = tmp_path / 'irac_ch1_flight.fits'
        hdulist.writeto(filepath)

        with patch('photutils.datasets.load._get_path',
                   return_value=str(filepath)):
            hdu = load.load_irac_psf(1)

        assert isinstance(hdu, fits.ImageHDU)
        assert np.array_equal(hdu.data, data)
        assert hdu.header['TELESCOP'] == 'Spitzer'

    def test_load_irac_psf_all_channels(self, tmp_path):
        """
        Test the load_irac_psf function for all valid channels.
        """
        for channel in range(1, 5):
            data = np.random.default_rng(seed=0).random((41, 41))
            header = fits.Header()
            header['CHNLNUM'] = channel

            primary_hdu = fits.PrimaryHDU(data=data, header=header)
            hdulist = fits.HDUList([primary_hdu])

            filepath = tmp_path / f'irac_ch{channel}_flight.fits'
            hdulist.writeto(filepath, overwrite=True)

            with patch('photutils.datasets.load._get_path',
                       return_value=str(filepath)):
                hdu = load.load_irac_psf(channel)

            assert isinstance(hdu, fits.ImageHDU)
            assert hdu.header['CHNLNUM'] == channel

    def test_load_irac_psf_show_progress(self, tmp_path):
        """
        Test the load_irac_psf function with show_progress=True.
        """
        data = np.random.default_rng(seed=0).random((41, 41))
        header = fits.Header()

        primary_hdu = fits.PrimaryHDU(data=data, header=header)
        hdulist = fits.HDUList([primary_hdu])

        filepath = tmp_path / 'irac_ch2_flight.fits'
        hdulist.writeto(filepath)

        with (
            patch('photutils.datasets.load._get_path',
                  return_value=str(filepath)) as mock_get_path,
        ):
            hdu = load.load_irac_psf(2, show_progress=True)

        assert isinstance(hdu, fits.ImageHDU)
        mock_get_path.assert_called_once_with(
            'irac_ch2_flight.fits',
            location='remote',
            show_progress=True,
        )

    def test_load_star_image_mocked(self, tmp_path):
        """
        Test the load_star_image function with mocked file download.
        """
        data = np.random.default_rng(seed=0).random((200, 200))
        header = fits.Header()
        header['OBJECT'] = 'M67'
        header['TELESCOP'] = 'Palomar'

        primary_hdu = fits.PrimaryHDU(data=data, header=header)
        hdulist = fits.HDUList([primary_hdu])

        filepath = tmp_path / 'M6707HH.fits'
        hdulist.writeto(filepath)

        with (pytest.warns(AstropyDeprecationWarning),
              patch('photutils.datasets.load._get_path',
                    return_value=str(filepath))):
            hdu = load.load_star_image()

        assert isinstance(hdu, fits.ImageHDU)
        assert np.array_equal(hdu.data, data)
        assert hdu.header['OBJECT'] == 'M67'

    def test_load_star_image_show_progress(self, tmp_path):
        """
        Test the load_star_image function with show_progress=True.
        """
        data = np.random.default_rng(seed=0).random((200, 200))
        header = fits.Header()

        primary_hdu = fits.PrimaryHDU(data=data, header=header)
        hdulist = fits.HDUList([primary_hdu])

        filepath = tmp_path / 'M6707HH.fits'
        hdulist.writeto(filepath)

        with (pytest.warns(AstropyDeprecationWarning),
              patch('photutils.datasets.load._get_path',
                    return_value=str(filepath)) as mock_get_path):
            hdu = load.load_star_image(show_progress=True)

        assert isinstance(hdu, fits.ImageHDU)
        mock_get_path.assert_called_once_with(
            'M6707HH.fits',
            location='remote',
            show_progress=True,
        )

    def test_load_simulated_hst_star_image(self, tmp_path):
        """
        Test the load_simulated_hst_star_image function with mocked file
        download.
        """
        data = np.random.default_rng(seed=0).random((300, 300))
        header = fits.Header()
        header['TELESCOP'] = 'HST'
        header['INSTRUME'] = 'WFC3'
        header['FILTER'] = 'F160W'

        primary_hdu = fits.PrimaryHDU(data=data, header=header)
        hdulist = fits.HDUList([primary_hdu])

        filepath = tmp_path / 'hst_wfc3ir_f160w_simulated_starfield.fits'
        hdulist.writeto(filepath)

        with patch('photutils.datasets.load._get_path',
                   return_value=str(filepath)):
            hdu = load.load_simulated_hst_star_image()

        assert isinstance(hdu, fits.ImageHDU)
        assert np.array_equal(hdu.data, data)
        assert hdu.header['TELESCOP'] == 'HST'
        assert hdu.header['INSTRUME'] == 'WFC3'
        assert hdu.header['FILTER'] == 'F160W'

    def test_load_simulated_hst_star_image_show_progress(self, tmp_path):
        """
        Test the load_simulated_hst_star_image function with
        show_progress=True.
        """
        data = np.random.default_rng(seed=0).random((300, 300))
        header = fits.Header()

        primary_hdu = fits.PrimaryHDU(data=data, header=header)
        hdulist = fits.HDUList([primary_hdu])

        filepath = tmp_path / 'hst_wfc3ir_f160w_simulated_starfield.fits'
        hdulist.writeto(filepath)

        with (
            patch('photutils.datasets.load._get_path',
                  return_value=str(filepath)) as mock_get_path,
        ):
            hdu = load.load_simulated_hst_star_image(show_progress=True)

        assert isinstance(hdu, fits.ImageHDU)
        mock_get_path.assert_called_once_with(
            'hst_wfc3ir_f160w_simulated_starfield.fits',
            location='photutils-datasets',
            show_progress=True,
        )


def test_deprecated_get_path():
    """
    Test that the deprecated get_path function raises a warning and
    returns the expected path.
    """
    fn = '4gaussians_params.ecsv'
    with pytest.warns(AstropyDeprecationWarning):
        result = load.get_path(fn, location='local')
    assert fn in result



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the peakfinder module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.tests.helper import assert_quantity_allclose
from numpy.testing import assert_array_equal, assert_equal

from photutils.centroids import centroid_com
from photutils.datasets import make_gwcs, make_wcs
from photutils.detection import find_peaks
from photutils.utils._optional_deps import HAS_GWCS
from photutils.utils.exceptions import NoDetectionsWarning


class TestFindPeaks:
    def test_box_size(self, data):
        """
        Test with box_size.
        """
        tbl = find_peaks(data, 0.1, box_size=3)
        assert tbl['id'][0] == 1
        assert len(tbl) == 25
        columns = ['id', 'x_peak', 'y_peak', 'peak_value']
        assert all(column in tbl.colnames for column in columns)
        assert np.min(tbl['x_peak']) > 0
        assert np.max(tbl['x_peak']) < 101
        assert np.min(tbl['y_peak']) > 0
        assert np.max(tbl['y_peak']) < 101
        assert np.max(tbl['peak_value']) < 13.2

        # Test with units
        unit = u.Jy
        tbl2 = find_peaks(data << unit, 0.1 << unit, box_size=3)
        columns = ['id', 'x_peak', 'y_peak']
        for column in columns:
            assert_equal(tbl[column], tbl2[column])
        col = 'peak_value'
        assert tbl2[col].unit == unit
        assert_equal(tbl[col], tbl2[col].value)

    def test_footprint(self, data):
        """
        Test with footprint.
        """
        tbl0 = find_peaks(data, 0.1, box_size=3)
        tbl1 = find_peaks(data, 0.1, footprint=np.ones((3, 3)))
        assert_array_equal(tbl0, tbl1)

    def test_mask(self, data):
        """
        Test with mask.
        """
        mask = np.zeros(data.shape, dtype=bool)
        mask[0:50, :] = True
        tbl0 = find_peaks(data, 0.1, box_size=3)
        tbl1 = find_peaks(data, 0.1, box_size=3, mask=mask)
        assert len(tbl1) < len(tbl0)

    def test_mask_int(self, data):
        """
        Test that an integer mask gives the same result as a boolean
        mask.
        """
        bool_mask = np.zeros(data.shape, dtype=bool)
        bool_mask[0:50, :] = True
        int_mask = bool_mask.astype(int)

        tbl_bool = find_peaks(data, 0.1, box_size=3, mask=bool_mask)
        tbl_int = find_peaks(data, 0.1, box_size=3, mask=int_mask)
        assert_array_equal(tbl_bool, tbl_int)

    def test_maskshape(self, data):
        """
        Test if mask shape doesn't match data shape.
        """
        match = 'data and mask must have the same shape'
        with pytest.raises(ValueError, match=match):
            find_peaks(data, 0.1, mask=np.ones((5, 5)))

    def test_thresholdshape(self, data):
        """
        Test if threshold shape doesn't match data shape.
        """
        match = 'threshold array must have the same shape as the input data'
        with pytest.raises(ValueError, match=match):
            find_peaks(data, np.ones((2, 2)))

    def test_n_peaks(self, data):
        """
        Test n_peaks.
        """
        tbl = find_peaks(data, 0.1, box_size=3, n_peaks=1)
        assert len(tbl) == 1

    def test_border_width(self, data):
        """
        Test border exclusion.
        """
        tbl0 = find_peaks(data, 0.1, box_size=3)
        tbl1 = find_peaks(data, 0.1, box_size=3, border_width=0)
        tbl2 = find_peaks(data, 0.1, box_size=3, border_width=(0, 0))
        assert len(tbl0) == len(tbl1)
        assert len(tbl1) == len(tbl2)

        tbl3 = find_peaks(data, 0.1, box_size=3, border_width=25)
        tbl4 = find_peaks(data, 0.1, box_size=3, border_width=(25, 25))
        assert len(tbl3) == len(tbl4)
        assert len(tbl3) < len(tbl0)

        tbl0 = find_peaks(data, 0.1, box_size=3, border_width=(34, 0))
        tbl1 = find_peaks(data, 0.1, box_size=3, border_width=(0, 36))
        assert np.min(tbl0['y_peak']) >= 34
        assert np.min(tbl1['x_peak']) >= 36

        match = 'border_width must be >= 0'
        with pytest.raises(ValueError, match=match):
            find_peaks(data, 0.1, box_size=3, border_width=-1)
        match = 'border_width must have integer values'
        with pytest.raises(ValueError, match=match):
            find_peaks(data, 0.1, box_size=3, border_width=3.1)

    def test_border_width_excludes_all(self, data):
        """
        Test that a border_width encompassing the entire image returns
        None with a NoDetectionsWarning.
        """
        match = 'No local peaks were found'
        with pytest.warns(NoDetectionsWarning, match=match):
            tbl = find_peaks(data, 0.1, box_size=3, border_width=100)
        assert tbl is None

    def test_box_size_int(self, data):
        """
        Test noninteger box_size.
        """
        tbl1 = find_peaks(data, 0.1, box_size=5.0)
        tbl2 = find_peaks(data, 0.1, box_size=5.5)
        assert_array_equal(tbl1, tbl2)

    def test_centroid_func_callable(self, data):
        """
        Test that centroid_func is callable.
        """
        match = 'centroid_func must be a callable object'
        with pytest.raises(TypeError, match=match):
            find_peaks(data, 0.1, box_size=2, centroid_func=True)

    def test_centroid_func_with_error(self, data):
        """
        Test find_peaks with a centroid_func and an error array.
        """
        error = np.ones_like(data) * 0.1
        tbl = find_peaks(data, 0.1, box_size=3, centroid_func=centroid_com,
                         error=error)
        assert 'x_centroid' in tbl.colnames
        assert 'y_centroid' in tbl.colnames
        assert len(tbl) > 0

    def test_centroid_func_with_footprint(self, data):
        """
        Test find_peaks with a footprint and centroid_func.

        Even-sized footprint dimensions should be rounded up to odd for
        the centroid box_size.
        """
        footprint = np.ones((4, 4), dtype=bool)
        tbl = find_peaks(data, 0.1, footprint=footprint,
                         centroid_func=centroid_com)
        assert 'x_centroid' in tbl.colnames
        assert 'y_centroid' in tbl.colnames
        assert len(tbl) > 0

    def test_error_without_centroid_func(self, data):
        """
        Test that error is silently ignored when centroid_func is None.
        """
        error = np.ones_like(data) * 0.1
        tbl1 = find_peaks(data, 0.1, box_size=3)
        tbl2 = find_peaks(data, 0.1, box_size=3, error=error)
        assert_array_equal(tbl1, tbl2)

    def test_wcs(self, data):
        """
        Test with astropy WCS.
        """
        columns = ['skycoord_peak', 'skycoord_centroid']

        fits_wcs = make_wcs(data.shape)
        tbl = find_peaks(data, 1, wcs=fits_wcs, centroid_func=centroid_com)
        for column in columns:
            assert column in tbl.colnames
        assert tbl.colnames == ['id', 'x_peak', 'y_peak', 'skycoord_peak',
                                'peak_value', 'x_centroid', 'y_centroid',
                                'skycoord_centroid']

    @pytest.mark.skipif(not HAS_GWCS, reason='gwcs is required')
    def test_gwcs(self, data):
        """
        Test with gwcs.
        """
        columns = ['skycoord_peak', 'skycoord_centroid']

        gwcs_obj = make_gwcs(data.shape)
        tbl = find_peaks(data, 1, wcs=gwcs_obj, centroid_func=centroid_com)
        for column in columns:
            assert column in tbl.colnames

    @pytest.mark.skipif(not HAS_GWCS, reason='gwcs is required')
    def test_wcs_values(self, data):
        """
        Test that WCS and GWCS give the same sky coordinates.
        """
        fits_wcs = make_wcs(data.shape)
        gwcs_obj = make_gwcs(data.shape)
        tbl1 = find_peaks(data, 1, wcs=fits_wcs, centroid_func=centroid_com)
        tbl2 = find_peaks(data, 1, wcs=gwcs_obj, centroid_func=centroid_com)
        columns = ['skycoord_peak', 'skycoord_centroid']
        for column in columns:
            assert_quantity_allclose(tbl1[column].ra, tbl2[column].ra)
            assert_quantity_allclose(tbl1[column].dec, tbl2[column].dec)

    def test_constant_array(self):
        """
        Test for empty output table when data is constant.
        """
        data = np.ones((10, 10))
        match = 'Input data is constant'
        with pytest.warns(NoDetectionsWarning, match=match):
            tbl = find_peaks(data, 0.0)
        assert tbl is None

    def test_no_peaks(self, data):
        """
        Tests for when no peaks are found.
        """
        fits_wcs = make_wcs(data.shape)

        match = 'No local peaks were found'
        with pytest.warns(NoDetectionsWarning, match=match):
            tbl = find_peaks(data, 10000)
        assert tbl is None

        with pytest.warns(NoDetectionsWarning, match=match):
            tbl = find_peaks(data, 100000, centroid_func=centroid_com)
        assert tbl is None

        with pytest.warns(NoDetectionsWarning, match=match):
            tbl = find_peaks(data, 100000, wcs=fits_wcs)
        assert tbl is None

        with pytest.warns(NoDetectionsWarning, match=match):
            tbl = find_peaks(data, 100000, wcs=fits_wcs,
                             centroid_func=centroid_com)
        assert tbl is None

    def test_data_nans(self, data):
        """
        Test that data with NaNs does not issue Runtime warning.
        """
        data = np.copy(data)
        data[50:, :] = np.nan
        find_peaks(data, 0.1)

    def test_data_not_mutated(self, data):
        """
        Test that input data is not mutated by find_peaks.
        """
        data_copy = data.copy()
        find_peaks(data, 0.1, box_size=3)
        assert_equal(data, data_copy)

    def test_data_not_mutated_with_mask(self, data):
        """
        Test that input data is not mutated when a mask is used.
        """
        data_copy = data.copy()
        mask = np.zeros(data.shape, dtype=bool)
        mask[0:50] = True
        find_peaks(data, 0.1, box_size=3, mask=mask)
        assert_equal(data, data_copy)

    @pytest.mark.parametrize('box_size', [3, 5, 7, 11])
    def test_box_size_min_separation(self, box_size):
        """
        Test that box_size imposes a minimum separation of ``box_size //
        2 + 1`` pixels between detected peaks.
        """
        min_sep = box_size // 2 + 1
        size = 10 * box_size
        img = np.zeros((size, size))

        # Place two peaks exactly at the minimum allowed separation
        cy = size // 2
        cx1 = size // 2
        cx2 = cx1 + min_sep
        img[cy, cx1] = 10.0
        img[cy, cx2] = 10.0

        tbl = find_peaks(img, 1.0, box_size=box_size)
        assert len(tbl) == 2

    @pytest.mark.parametrize('box_size', [3, 5, 7, 11])
    def test_box_size_below_min_separation(self, box_size):
        """
        Test that peaks separated by less than ``box_size // 2 + 1``
        pixels are merged (only the brighter one survives).
        """
        min_sep = box_size // 2 + 1
        size = 10 * box_size
        img = np.zeros((size, size))

        # Place two peaks one pixel closer than the minimum separation;
        # only the brighter peak should survive
        cy = size // 2
        cx1 = size // 2
        cx2 = cx1 + min_sep - 1
        img[cy, cx1] = 10.0
        img[cy, cx2] = 8.0

        tbl = find_peaks(img, 1.0, box_size=box_size)
        assert len(tbl) == 1
        assert tbl['peak_value'][0] == 10.0

    def test_min_separation(self, data):
        """
        Test that min_separation enforces minimum Euclidean distance.
        """
        tbl0 = find_peaks(data, 0.1, box_size=3)
        tbl1 = find_peaks(data, 0.1, box_size=3, min_separation=10)
        assert len(tbl1) <= len(tbl0)

        # Check that all pairs of peaks are at least min_separation
        # apart
        if len(tbl1) > 1:
            x = np.array(tbl1['x_peak'], dtype=float)
            y = np.array(tbl1['y_peak'], dtype=float)
            for i in range(len(tbl1)):
                for j in range(i + 1, len(tbl1)):
                    dist = np.sqrt((x[i] - x[j])**2 + (y[i] - y[j])**2)
                    assert dist > 10

    def test_min_separation_two_peaks(self):
        """
        Test min_separation with two peaks at known separation.
        """
        data = np.zeros((100, 100))
        data[50, 30] = 10.0
        data[50, 50] = 8.0

        # Separation is 20 pixels; min_separation=15 should keep both
        tbl = find_peaks(data, 1.0, box_size=3, min_separation=15)
        assert len(tbl) == 2

        # min_separation=25 should keep only the brightest
        tbl = find_peaks(data, 1.0, box_size=3, min_separation=25)
        assert len(tbl) == 1
        assert tbl['peak_value'][0] == 10.0

    def test_min_separation_plateau(self):
        """
        Test that min_separation treats plateaus identically to an
        equivalent circular footprint (all equal-valued plateau pixels
        are local maxima).
        """
        data = np.zeros((50, 50))
        data[20:30, 20:30] = 10.0  # 10x10 plateau (diagonal ~12.7 px)

        for radius in (5, 15):
            idx = np.arange(-radius, radius + 1)
            xx, yy = np.meshgrid(idx, idx)
            fp = np.array((xx**2 + yy**2) <= radius**2, dtype=int)
            tbl_ref = find_peaks(data, 1.0, footprint=fp)
            tbl_fast = find_peaks(data, 1.0, min_separation=radius)

            assert len(tbl_ref) == len(tbl_fast)
            assert_array_equal(tbl_ref['x_peak'], tbl_fast['x_peak'])
            assert_array_equal(tbl_ref['y_peak'], tbl_fast['y_peak'])

    def test_min_separation_with_units(self):
        """
        Test min_separation with Quantity data.
        """
        unit = u.Jy
        data = np.zeros((100, 100))
        data[50, 30] = 10.0
        data[50, 50] = 8.0

        tbl = find_peaks(data << unit, 1.0 << unit, box_size=3,
                         min_separation=25)
        assert len(tbl) == 1
        assert tbl['peak_value'][0].value == 10.0
        assert tbl['peak_value'][0].unit == unit

    def test_min_separation_with_npeaks(self):
        """
        Test that min_separation and n_peaks work together.
        """
        data = np.zeros((100, 100))
        data[20, 20] = 10.0
        data[20, 60] = 8.0
        data[60, 20] = 6.0
        data[60, 60] = 4.0

        # All peaks are well-separated;
        # n_peaks=2 should keep brightest 2
        tbl = find_peaks(data, 1.0, min_separation=5, n_peaks=2)
        assert len(tbl) == 2
        assert tbl['peak_value'][0] == 10.0
        assert tbl['peak_value'][1] == 8.0

    def test_min_separation_negative(self, data):
        """
        Test that negative min_separation raises ValueError.
        """
        match = 'min_separation must be >= 0'
        with pytest.raises(ValueError, match=match):
            find_peaks(data, 0.1, min_separation=-1)

    def test_min_separation_zero(self, data):
        """
        Test that min_separation=0 gives the same result as None.
        """
        tbl0 = find_peaks(data, 0.1, box_size=3)
        tbl1 = find_peaks(data, 0.1, box_size=3, min_separation=0)
        assert_array_equal(tbl0, tbl1)

    def test_min_separation_with_footprint(self, data):
        """
        Test that min_separation takes priority over footprint.
        """
        footprint = np.ones((3, 3))
        tbl = find_peaks(data, 0.1, footprint=footprint, min_separation=10)
        assert len(tbl) > 0

        # Check minimum separation is enforced
        if len(tbl) > 1:
            x = np.array(tbl['x_peak'], dtype=float)
            y = np.array(tbl['y_peak'], dtype=float)
            for i in range(len(tbl)):
                for j in range(i + 1, len(tbl)):
                    dist = np.sqrt((x[i] - x[j])**2 + (y[i] - y[j])**2)
                    assert dist > 10

    def test_min_separation_matches_circular_footprint(self):
        """
        Test that min_separation produces the same peaks as an
        equivalent circular footprint passed to maximum_filter.
        """
        rng = np.random.default_rng(42)
        data = rng.standard_normal((200, 200))
        data[50, 50] = 20.0
        data[120, 130] = 18.0
        data[30, 170] = 15.0
        threshold = 3.0

        for radius in (5, 10, 25, 50):
            # Reference: actual circular footprint (slow but correct)
            idx = np.arange(-radius, radius + 1)
            xx, yy = np.meshgrid(idx, idx)
            fp = np.array((xx**2 + yy**2) <= radius**2, dtype=int)
            tbl_ref = find_peaks(data, threshold, footprint=fp)

            tbl_fast = find_peaks(data, threshold, min_separation=radius)

            if tbl_ref is None:
                assert tbl_fast is None
            else:
                ref_xy = set(zip(tbl_ref['x_peak'].tolist(),
                                 tbl_ref['y_peak'].tolist(),
                                 strict=True))
                fast_xy = set(zip(tbl_fast['x_peak'].tolist(),
                                  tbl_fast['y_peak'].tolist(),
                                  strict=True))
                assert ref_xy == fast_xy

    def test_min_separation_rejects_non_maxima(self):
        """
        Test that min_separation rejects peaks that are not the true
        local maximum within the circular region.

        This test would fail with a greedy KD-tree-only approach that
        uses a small box_size for initial peak detection, because such
        an approach would keep faint peaks that are not the maximum
        within a circle of min_separation (due to non-peak pixels with
        higher values in the neighborhood).
        """
        data = np.zeros((100, 100))

        # Bright peak with a declining gradient
        data[50, 50] = 100.0
        for i in range(1, 30):
            data[50, 50 + i] = 100.0 - 2 * i  # 98, 96, ..., 42

        # Faint peak at (50, 85), which is 35 px from the bright peak.
        # The gradient pixel at (50, 65) = 100 - 2*15 = 70, which is
        # within radius 20 of (50, 85) and brighter (70 > 45).
        data[50, 85] = 45.0

        # With min_separation=20: (50, 85) is NOT the local max within a
        # circle of radius 20 because (50, 65)=70 > 45.
        tbl = find_peaks(data, 1.0, min_separation=20)
        assert len(tbl) == 1
        assert tbl['x_peak'][0] == 50
        assert tbl['y_peak'][0] == 50

    def test_min_separation_keeps_nearby_true_maxima(self):
        """
        Test that two equal-valued peaks within min_separation of each
        other are both retained, matching the circular footprint result.
        """
        radius = 12
        data = np.zeros((100, 100))

        # Two equal-valued peaks separated by less than min_separation
        # (dist = 11 px < radius = 12 px). Because they are equal,
        # each is tied for the max in its own circle, so both should
        # be detected (same behavior as maximum_filter with a circular
        # footprint).
        data[50, 40] = 10.0
        data[50, 51] = 10.0

        # Reference: actual circular footprint
        idx = np.arange(-radius, radius + 1)
        xx, yy = np.meshgrid(idx, idx)
        fp = np.array((xx**2 + yy**2) <= radius**2, dtype=int)
        tbl_ref = find_peaks(data, 1.0, footprint=fp)
        tbl_fast = find_peaks(data, 1.0, min_separation=radius)

        assert len(tbl_ref) == 2
        assert len(tbl_fast) == 2
        assert_array_equal(tbl_ref['x_peak'], tbl_fast['x_peak'])
        assert_array_equal(tbl_ref['y_peak'], tbl_fast['y_peak'])

    def test_nan_no_false_peaks(self):
        """
        Test that NaN pixels do not produce false peaks when the fill
        value (nanmin) happens to be a local maximum.
        """
        data = np.full((50, 50), 5.0)
        data[25, 25] = 10.0  # one real peak
        data[10, 10] = np.nan  # NaN pixel (fill value = 5.0 = background)
        data[10, 11] = np.nan

        tbl = find_peaks(data, 6.0, box_size=3)
        assert len(tbl) == 1
        assert tbl['x_peak'][0] == 25
        assert tbl['y_peak'][0] == 25

    def test_nan_adjacent_to_peak(self):
        """
        Test that NaN pixels adjacent to a real peak do not cause the
        peak to be lost or duplicated.
        """
        data = np.zeros((50, 50))
        data[25, 25] = 10.0
        data[25, 26] = np.nan
        data[24, 25] = np.nan

        tbl = find_peaks(data, 1.0, box_size=3)
        assert len(tbl) == 1
        assert tbl['x_peak'][0] == 25
        assert tbl['y_peak'][0] == 25

    def test_all_negative_data(self):
        """
        Test peak detection with all-negative data.

        Peaks near the border may be suppressed because maximum_filter
        uses cval=0.0, but interior peaks should be detected correctly.
        """
        data = np.full((50, 50), -10.0)
        data[25, 25] = -1.0  # brightest pixel, well inside border

        tbl = find_peaks(data, -5.0, box_size=3)
        assert tbl is not None
        assert len(tbl) == 1
        assert tbl['x_peak'][0] == 25
        assert tbl['y_peak'][0] == 25

    def test_all_negative_border_suppression(self):
        """
        Test that all-negative data near the border is suppressed by
        cval=0.0 in maximum_filter.
        """
        data = np.full((50, 50), -10.0)
        # Peak at border and one well inside
        data[0, 0] = -1.0
        data[25, 25] = -1.0

        # The peak at (0,0) is above the threshold but cval=0.0 means
        # the border region has a "virtual" maximum of 0.0, which is
        # greater than -1.0, so this pixel won't be detected as a peak.
        tbl = find_peaks(data, -5.0, box_size=3)
        assert tbl is not None
        # The border peak should not be among the results
        assert not any((tbl['x_peak'] == 0) & (tbl['y_peak'] == 0))
        # The interior peak should be detected
        assert any((tbl['x_peak'] == 25) & (tbl['y_peak'] == 25))

    def test_min_separation_with_centroid_func(self, data):
        """
        Test that min_separation works with centroid_func.

        The centroid box_size defaults to box_size (3) when
        min_separation is used.
        """
        tbl = find_peaks(data, 0.1, min_separation=10,
                         centroid_func=centroid_com)
        assert 'x_centroid' in tbl.colnames
        assert 'y_centroid' in tbl.colnames
        assert len(tbl) > 0



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for measuring morphological properties of sources.
"""

import numpy as np

from photutils.utils._deprecation import deprecated_positional_kwargs

__all__ = ['data_properties']


@deprecated_positional_kwargs(since='3.0', until='4.0')
def data_properties(data, mask=None, background=None, wcs=None):
    """
    Calculate the morphological properties (and centroid) of a 2D array
    (e.g., an image cutout of an object) using image moments.

    Parameters
    ----------
    data : array_like or `~astropy.units.Quantity`
        The 2D array of the image.

    mask : array_like (bool), optional
        A boolean mask, with the same shape as ``data``, where a `True`
        value indicates the corresponding element of ``data`` is masked.
        Masked data are excluded from all calculations.

    background : float or array_like, optional
        The background level previously present in the input ``data``.
        ``background`` may be a scalar value or a 2D array with the same
        shape as ``data``. The input ``background`` is not subtracted
        from ``data``, which should already be background-subtracted;
        providing it only enables background-related properties to be
        measured.

    wcs : WCS object or `None`, optional
        A world coordinate system (WCS) transformation that
        supports the `astropy shared interface for WCS
        <https://docs.astropy.org/en/stable/wcs/wcsapi.html>`_ (e.g.,
        `astropy.wcs.WCS`, `gwcs.wcs.WCS`). If `None`, then all
        sky-based properties will be set to `None`.

    Returns
    -------
    result : `~photutils.segmentation.SourceCatalog` instance
        A scalar `~photutils.segmentation.SourceCatalog` object (single
        source) containing the morphological properties.

    Raises
    ------
    ValueError
        If ``data`` is not a 2D array.

    ValueError
        If ``mask`` is provided and does not have the same shape as
        ``data``.

    ValueError
        If ``mask`` is provided and all pixels are masked.

    ValueError
        If ``background`` is provided and is not a scalar or a 2D array
        with the same shape as ``data``.
    """
    # Prevent circular import
    from photutils.segmentation import SegmentationImage, SourceCatalog

    data = np.asanyarray(data)
    if len(data.shape) != 2:
        msg = 'data must be a 2D array'
        raise ValueError(msg)

    seg_arr = np.ones(data.shape, dtype=int)
    if mask is not None:
        mask = np.asarray(mask, dtype=bool)
        if mask.shape != data.shape:
            msg = 'mask must have the same shape as data'
            raise ValueError(msg)
        if np.all(mask):
            msg = 'All pixels in data are masked'
            raise ValueError(msg)
        seg_arr[mask] = 0

    segment_image = SegmentationImage(seg_arr)

    if background is not None:
        background = np.asarray(background)
        if background.ndim == 0:
            background = np.full(data.shape, float(background))
        elif background.shape != data.shape:
            msg = ('background must be a scalar or a 2D array '
                   'with the same shape as data')
            raise ValueError(msg)

    # mask is encoded in seg_arr (masked pixels set to 0), so
    # mask=None is intentional here
    return SourceCatalog(data, segment_image, mask=None,
                         background=background, wcs=wcs)[0]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the simulation module.
"""

import numpy as np
import pytest
from astropy.modeling.models import Gaussian2D
from astropy.table import Table
from numpy.testing import assert_equal

from photutils.psf import (CircularGaussianPRF, make_psf_model,
                           make_psf_model_image)


def test_make_psf_model_image():
    shape = (401, 451)
    n_sources = 100
    model = CircularGaussianPRF(fwhm=2.7)
    data, params = make_psf_model_image(shape, model, n_sources)
    assert data.shape == shape
    assert isinstance(params, Table)
    assert len(params) == n_sources

    model_shape = (13, 13)
    data2, params2 = make_psf_model_image(shape, model, n_sources,
                                          model_shape=model_shape)
    assert_equal(data, data2)
    assert len(params2) == n_sources

    flux = (100, 200)
    fwhm = (2.5, 4.5)
    alpha = (0, 1)
    n_sources = 10
    data, params = make_psf_model_image(shape, model, n_sources,
                                        seed=0, flux=flux, fwhm=fwhm,
                                        alpha=alpha)
    assert len(params) == n_sources
    colnames = ('id', 'x_0', 'y_0', 'flux', 'fwhm')
    for colname in colnames:
        assert colname in params.colnames
    assert 'alpha' not in params.colnames
    assert np.min(params['flux']) >= flux[0]
    assert np.max(params['flux']) <= flux[1]
    assert np.min(params['fwhm']) >= fwhm[0]
    assert np.max(params['fwhm']) <= fwhm[1]


def test_make_psf_model_image_custom():
    shape = (401, 451)
    n_sources = 100
    model = Gaussian2D()
    psf_model = make_psf_model(model, x_name='x_mean', y_name='y_mean')
    data, params = make_psf_model_image(shape, psf_model, n_sources,
                                        model_shape=(11, 11))
    assert data.shape == shape
    assert isinstance(params, Table)
    assert len(params) == n_sources


def test_make_psf_model_image_inputs():
    shape = (50, 50)
    match = 'psf_model must be an Astropy Model subclass'
    with pytest.raises(TypeError, match=match):
        make_psf_model_image(shape, None, 2)

    match = 'psf_model must be two-dimensional'
    model = CircularGaussianPRF(fwhm=2.7)
    model.n_inputs = 3
    with pytest.raises(ValueError, match=match):
        make_psf_model_image(shape, model, 2)

    match = 'model_shape must be specified if the model does not have'
    model = CircularGaussianPRF(fwhm=2.7)
    model.bounding_box = None
    with pytest.raises(ValueError, match=match):
        make_psf_model_image(shape, model, 2)

    match = 'Invalid PSF model - could not find PSF parameter names'
    model = Gaussian2D()
    with pytest.raises(ValueError, match=match):
        make_psf_model_image(shape, model, 2)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the mask module.
"""

import astropy.units as u
import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_almost_equal

from photutils.aperture.bounding_box import BoundingBox
from photutils.aperture.circle import CircularAnnulus, CircularAperture
from photutils.aperture.mask import ApertureMask
from photutils.aperture.rectangle import RectangularAnnulus

POSITIONS = [(-20, -20), (-20, 20), (20, -20), (60, 60)]


def test_mask_input_shapes():
    mask_data = np.ones((10, 10))
    bbox = BoundingBox(5, 10, 5, 10)
    match = 'mask data and bounding box must have the same shape'
    with pytest.raises(ValueError, match=match):
        ApertureMask(mask_data, bbox)


def test_mask_array():
    mask_data = np.ones((10, 10))
    bbox = BoundingBox(5, 15, 5, 15)
    mask = ApertureMask(mask_data, bbox)
    data = np.array(mask)
    assert_allclose(data, mask.data)


def test_mask_copy():
    bbox = BoundingBox(5, 15, 5, 15)

    mask = ApertureMask(np.ones((10, 10)), bbox)
    mask_copy = np.array(mask, copy=True)
    mask_copy[0, 0] = 100.0
    assert mask.data[0, 0] == 1.0

    mask = ApertureMask(np.ones((10, 10)), bbox)
    mask_copy = np.array(mask, copy=False)
    mask_copy[0, 0] = 100.0
    assert mask.data[0, 0] == 100.0

    # No copy: copy=None returns a copy only if __array__ returns a
    # copy; copy=None was introduced in NumPy 2.0
    mask = ApertureMask(np.ones((10, 10)), bbox)
    mask_copy = np.array(mask, copy=None)
    mask_copy[0, 0] = 100.0
    assert mask.data[0, 0] == 100.0

    # No copy
    mask = ApertureMask(np.ones((10, 10)), bbox)
    mask_copy = np.asarray(mask)
    mask_copy[0, 0] = 100.0
    assert mask.data[0, 0] == 100.0

    # Needs to copy because of the dtype change
    mask = ApertureMask(np.ones((10, 10)), bbox)
    mask_copy = np.asarray(mask, dtype=int)
    mask_copy[0, 0] = 100.0
    assert mask.data[0, 0] == 1.0


def test_mask_get_overlap_slices():
    aper = CircularAperture((5, 5), r=10.0)
    mask = aper.to_mask()
    slc = ((slice(0, 16, None), slice(0, 16, None)),
           (slice(5, 21, None), slice(5, 21, None)))
    assert mask.get_overlap_slices((25, 25)) == slc


def test_mask_cutout_shape():
    mask_data = np.ones((10, 10))
    bbox = BoundingBox(5, 15, 5, 15)
    mask = ApertureMask(mask_data, bbox)

    match = 'data must be a 2D array'
    with pytest.raises(ValueError, match=match):
        mask.cutout(np.arange(10))

    match = 'input shape must have 2 elements'
    with pytest.raises(ValueError, match=match):
        mask.to_image((10,))


def test_mask_cutout_copy():
    data = np.ones((50, 50))
    aper = CircularAperture((25, 25), r=10.0)
    mask = aper.to_mask()
    cutout = mask.cutout(data, copy=True)
    data[25, 25] = 100.0
    assert cutout[10, 10] == 1.0

    # Test quantity data
    data2 = np.ones((50, 50)) * u.adu
    cutout2 = mask.cutout(data2, copy=True)
    assert cutout2.unit == data2.unit
    data2[25, 25] = 100.0 * u.adu
    assert cutout2[10, 10].value == 1.0


@pytest.mark.parametrize('position', POSITIONS)
def test_mask_cutout_no_overlap(position):
    data = np.ones((50, 50))
    aper = CircularAperture(position, r=10.0)
    mask = aper.to_mask()

    cutout = mask.cutout(data)
    assert cutout is None

    weighted_data = mask.multiply(data)
    assert weighted_data is None

    image = mask.to_image(data.shape)
    assert image is None


@pytest.mark.parametrize('position', POSITIONS)
def test_mask_cutout_partial_overlap(position):
    data = np.ones((50, 50))
    aper = CircularAperture(position, r=30.0)
    mask = aper.to_mask()

    cutout = mask.cutout(data)
    assert cutout.shape == mask.shape

    weighted_data = mask.multiply(data)
    assert weighted_data.shape == mask.shape

    image = mask.to_image(data.shape)
    assert image.shape == data.shape


def test_mask_cutout_partial_overlap_quantity():
    """
    Test that cutout with a Quantity array and partial overlap applies
    the data unit to the output cutout (covers the `cutout <<=
    data.unit` branch).
    """
    aper = CircularAperture((-20, -20), r=30.0)
    mask = aper.to_mask()
    data = np.ones((50, 50)) * u.adu
    cutout = mask.cutout(data)
    assert isinstance(cutout, u.Quantity)
    assert cutout.unit == u.adu


def test_mask_multiply():
    radius = 10.0
    data = np.ones((50, 50))
    aper = CircularAperture((25, 25), r=radius)
    mask = aper.to_mask()
    data_weighted = mask.multiply(data)
    assert_almost_equal(np.sum(data_weighted), np.pi * radius**2)

    # Test that multiply() returns a copy
    data[25, 25] = 100.0
    assert data_weighted[10, 10] == 1.0


def test_mask_multiply_quantity():
    radius = 10.0
    data = np.ones((50, 50)) * u.adu
    aper = CircularAperture((25, 25), r=radius)
    mask = aper.to_mask()
    data_weighted = mask.multiply(data)
    assert data_weighted.unit == u.adu
    assert_almost_equal(np.sum(data_weighted.value), np.pi * radius**2)

    # Test that multiply() returns a copy
    data[25, 25] = 100.0 * u.adu
    assert data_weighted[10, 10].value == 1.0


@pytest.mark.parametrize('value', [np.nan, np.inf])
def test_mask_nonfinite_fill_value(value):
    aper = CircularAnnulus((0, 0), 10, 20)
    data = np.ones((101, 101)).astype(int)
    cutout = aper.to_mask().cutout(data, fill_value=value)
    assert ~np.isfinite(cutout[0, 0])


def test_mask_multiply_fill_value():
    aper = CircularAnnulus((0, 0), 10, 20)
    data = np.ones((101, 101)).astype(int)
    cutout = aper.to_mask().multiply(data, fill_value=np.nan)
    xypos = ((20, 20), (5, 5), (5, 35), (35, 5), (35, 35))
    for x, y in xypos:
        assert np.isnan(cutout[y, x])


def test_mask_nonfinite_in_bbox():
    """
    Regression test that non-finite data values outside the mask but
    within the bounding box are set to zero.
    """
    data = np.ones((101, 101))
    data[33, 33] = np.nan
    data[67, 67] = np.inf
    data[33, 67] = -np.inf
    data[22, 22] = np.nan
    data[22, 23] = np.inf

    radius = 20.0
    aper1 = CircularAperture((50, 50), r=radius)
    aper2 = CircularAperture((5, 5), r=radius)

    wdata1 = aper1.to_mask(method='exact').multiply(data)
    assert_allclose(np.sum(wdata1), np.pi * radius**2)

    wdata2 = aper2.to_mask(method='exact').multiply(data)
    assert_allclose(np.sum(wdata2), 561.6040111923013)


def test_mask_get_values():
    aper = CircularAnnulus(((0, 0), (50, 50), (100, 100)), 10, 20)
    data = np.ones((101, 101))
    values = [mask.get_values(data) for mask in aper.to_mask()]
    shapes = [val.shape for val in values]
    sums = [np.sum(val) for val in values]
    assert shapes[0] == (278,)
    assert shapes[1] == (1068,)
    assert shapes[2] == (278,)
    sums_expected = (245.621534, 942.477796, 245.621534)
    assert_allclose(sums, sums_expected)


def test_mask_get_values_no_overlap():
    aper = CircularAperture((-100, -100), r=3)
    data = np.ones((51, 51))
    values = aper.to_mask().get_values(data)
    assert values.shape == (0,)


def test_mask_get_values_mask():
    aper = CircularAperture((24.5, 24.5), r=10.0)
    data = np.ones((51, 51))
    mask = aper.to_mask()
    match = 'mask and data must have the same shape'
    with pytest.raises(ValueError, match=match):
        mask.get_values(data, mask=np.ones(3))

    arr = mask.get_values(data, mask=None)
    assert_allclose(np.sum(arr), 100.0 * np.pi)

    data_mask = np.zeros(data.shape, dtype=bool)
    data_mask[25:] = True
    arr2 = mask.get_values(data, mask=data_mask)
    assert_allclose(np.sum(arr2), 100.0 * np.pi / 2.0)


def test_rectangular_annulus_hin():
    aper = RectangularAnnulus((25, 25), 2, 4, 20, h_in=18, theta=0)
    mask = aper.to_mask(method='center')
    assert mask.data.shape == (21, 5)
    assert np.count_nonzero(mask.data) == 40
